import numpy as np
from numpy.random import Generator
from scipy.optimize import fsolve
from numpy.typing import NDArray, ArrayLike
from typing import Any
from cratermaker.core.crater import Crater
from cratermaker.utils.custom_types import FloatLike
from cratermaker.components.surface import Surface
from cratermaker.components.morphology import Morphology
from cratermaker.utils.general_utils import parameter
from cratermaker._cratermaker import crater_functions, ejecta_functions, morphology_functions

@Morphology.register("simplemoon")
class SimpleMoon(Morphology):
    """
    An operations class for computing the morphology of a crater and applying it to a surface mesh.

    Parameters
    ----------
    ejecta_truncation : float, optional
        The relative distance from the rim of the crater to truncate the ejecta blanket, default is None, which will compute a 
        truncation distance based on where the ejecta thickness reaches a small value.         
    rng : Generator, optional
        A random number generator instance. If not provided, the default numpy RNG will be used.
    dorays : bool, optional
        A flag to determine if the ray pattern should be used instead of the homogeneous ejecta blanket, default is True.
    rng : numpy.random.Generator | None
        A numpy random number generator. If None, a new generator is created using the rng_seed if it is provided.
    rng_seed : Any type allowed by the rng_seed argument of numpy.random.Generator, optional
        The rng_rng_seed for the RNG. If None, a new RNG is created.
    rng_state : dict, optional
        The state of the random number generator. If None, a new state is created.
    **kwargs : Any
        Additional keyword arguments.
    """
    
    def __init__(self, 
                 ejecta_truncation: FloatLike | None = None,
                 dorays: bool = True,
                 rng: Generator | None = None,
                 rng_seed: int | None = None,
                 rng_state: dict | None = None,
                 **kwargs: Any 
                 ):
        super().__init__(rng=rng, rng_seed=rng_seed, rng_state=rng_state, **kwargs)
        object.__setattr__(self, "_ejecta_truncation" , None)
        object.__setattr__(self, "_rimheight" , None)
        object.__setattr__(self, "_rimwidth" , None)
        object.__setattr__(self, "_peakheight" , None)
        object.__setattr__(self, "_floor_diameter" , None)
        object.__setattr__(self, "_floordepth" , None)
        object.__setattr__(self, "_ejrim" , None)
        object.__setattr__(self, "_node" , None)

        self.ejecta_truncation = ejecta_truncation
        self.dorays = dorays

    def __repr__(self) -> str:
        base = super().__repr__()
        if self.ejecta_truncation is not None:
            base += f"\nEjecta Trunction: {self.ejecta_truncation:.2f} * crater.final_radius"
        else:
            base += f"\nEjecta Truncation: Off"
        return (
            f"{base}\n"
            f"Ejecta Rays: {self.dorays}"
        )

    def _set_morphology_config(self) -> None:
        """
        This method adds a crater to the morphology model and sets the parameters for the morphology based on the crater type. 
        """
        # Set the morphology based on crater type
        diameter_m = self.crater.final_diameter
        diameter_km = diameter_m * 1e-3  # Convert to km for some models

        if self.crater.morphology_type in ["simple", "transitional"]:
            # A hybrid model between Pike (1977) and Fassett & Thomson (2014)
            self.rimheight = 0.043 * diameter_km**1.014 * 1e3  # Closer to Fassett & Thomson
            self.rimwidth = 0.257 * diameter_km**1.011 * 1e3   # Pike model
            self.floordepth = 0.224 * diameter_km**1.010 * 1e3 # Closer to Fassett & Thomson
            self.floor_diameter = 0.200 * diameter_km**1.143 * 1e3  # Fassett & Thomson for D~1km, Pike for D~20km

        elif self.crater.morphology_type in ["complex", "peakring", "multiring"]:
            # Following Pike (1977)
            self.rimheight = 0.236 * diameter_km**0.399 * 1e3  # Pike model
            self.rimwidth = 0.467 * diameter_km**0.836 * 1e3   # Pike model
            self.floordepth = 1.044 * diameter_km**0.301 * 1e3 # Pike model
            # Fassett & Thomson for D~1km, Pike for D~20km, but limited to 90% of diameter
            self.floor_diameter = min(0.187 * diameter_km**1.249 * 1e3, 0.9 * diameter_m)
            self.peakheight = 0.032 * diameter_km**0.900 * 1e3  # Pike model
        else:
            raise ValueError(f"Unknown morphology type: {self.crater.morphology_type}")
            
        self.ejrim = 0.14 * (diameter_m * 0.5)**(0.74) # McGetchin et al. (1973) Thickness of ejecta at rim
        return


    def form_crater(self, 
                    surface: Surface,
                    crater: Crater | None = None, 
                    **kwargs) -> None:
        """
        This method forms the interior of the crater by altering the elevation variable of the surface mesh.
        
        Parameters
        ----------
        surface : Surface
            The surface to be altered.
        crater : Crater
            The crater object to be formed. This is optional if it has already been added
        **kwargs : dict
            Additional keyword arguments to be passed to internal functions (not used here).
        """
        self.crater = Crater.maker(crater, **kwargs)

        if not isinstance(surface, Surface):
            raise TypeError("surface must be an instance of Surface")
        self.node_index, self.face_index = surface.find_nearest_index(self.crater.location)

        # Test if the crater is big enough to modify the surface
        rmax = self.rmax(minimum_thickness=surface.smallest_length)
        region_view = surface.extract_region(self.crater.location, rmax)
        if region_view is None: # The crater is too small to change the surface
            return
        crater_area = np.pi * rmax**2
        
        # Check to make sure that the face at the crater location is not smaller than the crater area
        if surface.face_areas[self.face_index] > crater_area:
            return
        
        node_crater_distance, face_crater_distance = surface.get_distance(region_view, self.crater.location)
        reference_face_elevation, reference_node_elevation = surface.get_reference_surface(region_view, face_crater_distance, node_crater_distance, self.crater.location, self.crater.final_radius)
        
        try:
            node_elevation = self.crater_profile(node_crater_distance, 
                                                 reference_node_elevation)
            surface.node_elevation[region_view.node_indices] = node_elevation
            
            face_elevation = self.crater_profile(face_crater_distance, 
                                                 reference_face_elevation)
            surface.face_elevation[region_view.face_indices] = face_elevation
        except:
            print(self)
            raise ValueError("Something went wrong with this crater!")

        self.form_ejecta(surface, crater=self.crater, **kwargs) 
        return  


    def form_ejecta(self,
                    surface: Surface,
                    crater: Crater | None = None,
                    **kwargs) -> None:
        """
        This method forms the ejecta blanket around the crater by altering the elevation variable of the surface mesh.
       
        Parameters
        ----------
        surface : Surface
            The surface to be altered.
        **kwargs : dict
            Additional keyword arguments to be passed to internal functions (not used here). 
        """
        if crater:
            self.crater = crater

        if not isinstance(surface, Surface):
            raise TypeError("surface must be an instance of Surface")
        self.node_index, self.face_index = surface.find_nearest_index(self.crater.location) 

        # Test if the ejecta is big enough to modify the surface
        rmax = self.rmax(minimum_thickness=surface.smallest_length) 
        if not self.ejecta_truncation:
            ejecta_truncation = rmax / self.crater.final_radius
        else:
            ejecta_truncation = self.ejecta_truncation
        region_view = surface.extract_region(self.crater.location, rmax)
        if region_view is None: # The crater is too small to change the surface
            return
        ejecta_area = np.pi * rmax**2
        
        # Check to make sure that the face at the crater location is not smaller than the ejecta blanket area
        if surface.face_areas[self.face_index] > ejecta_area:
            return
        
        node_crater_distance, face_crater_distance = surface.get_distance(region_view, self.crater.location)
        node_crater_bearing, face_crater_bearing  = surface.get_initial_bearing(region_view, self.crater.location)
        morphology_functions.form_ejecta(self, 
                                         region_view, 
                                         node_crater_distance, 
                                         face_crater_distance, 
                                         node_crater_bearing,
                                         face_crater_bearing,
                                         ejecta_truncation,
                                         surface.node_elevation,
                                         surface.face_elevation,
                                         surface.ejecta_thickness,
                                         surface.ray_intensity)
        return


    def crater_profile(self, r: ArrayLike, r_ref: ArrayLike) -> NDArray[np.float64]:
        elevation = crater_functions.profile(r,
                                   r_ref, 
                                   self.crater.final_diameter, 
                                   self.floordepth, 
                                   self.floor_diameter, 
                                   self.rimheight, 
                                   self.ejrim
                                )
        
        return np.array(elevation, dtype=np.float64)
    

    def ejecta_profile(self, r: ArrayLike) -> NDArray[np.float64]:
        elevation = ejecta_functions.profile(r,
                                   self.crater.final_diameter, 
                                   self.ejrim
                                )
        elevation = np.array(elevation, dtype=np.float64)
        return elevation
   
    
    def ejecta_distribution(self, r: ArrayLike, theta: ArrayLike) -> NDArray[np.float64]:
        thickness = ejecta_functions.distribution(r, theta,
                                       self.crater.final_diameter, 
                                       self.ejrim, 
                                       self.ejecta_truncation,
                                       self.dorays
                                    )
        thickness = np.array(thickness, dtype=np.float64)
        return thickness


    def ray_intensity(self, r: ArrayLike, theta: ArrayLike) -> NDArray[np.float64]:
        intensity = ejecta_functions.ray_intensity(r, theta,
                                       self.crater.final_diameter, 
                                       self.ejecta_truncation,
                                    )
        intensity = np.array(intensity, dtype=np.float64)
        return intensity


    def rmax(self, 
            minimum_thickness: FloatLike,
            feature: str = "ejecta",
            crater: Crater | None = None) -> float:
        """
        Compute the maximum extent of the crater based on the minimum thickness of a feature, or the ejecta_truncation factor,
        whichever is smaller.
        
        Parameters
        ----------
        minimum_thickness : FloatLike
            The minimum thickness of the feature blanket in meters.
        feature : str, optional, default = "ejecta"
            The feature to compute the maximum extent. Either "crater" or "ejecta". If "crater" is chosen, the rmax is based
            on where the raised rim is smaller than minimum thickness. 
        crater : Crater, optional
            The crater object to be used. If None, the current crater object is used. If passed, the current crater object is replaced.
        Returns
        -------
        float
            The maximum extent of the crater or ejecta blanket in meters.
        """ 
        
        # Compute the reference surface for the crater 
        if crater is not None:
            self.crater = crater

        if feature == "ejecta":
            def _profile_invert(r):
                return self.ejecta_profile(r) - minimum_thickness
        elif feature == "crater":
            def _profile_invert(r):
                return self.crater_profile(r, np.zeros(1)) - minimum_thickness
        else:
            raise ValueError("Unknown feature type. Choose either 'crater' or 'ejecta'")
    
        # Get the maximum extent 
        rmax = fsolve(_profile_invert, x0=self.crater.final_radius*1.01)[0]
        
        if self.ejecta_truncation:
            rmax = min(rmax, self.ejecta_truncation * self.crater.final_radius)
            
        if feature == "crater":
            rmax = max(rmax, self.crater.final_radius)

        return float(rmax)

    @property
    def rimheight(self) -> float:
        """
        The height of the crater rim in meters.
        
        Returns
        -------
        float
        """
        return self._rimheight
    
    @rimheight.setter
    def rimheight(self, value: FloatLike) -> None:
        if not isinstance(value, FloatLike):
            raise TypeError("rimheight must be of type FloatLike")
        self._rimheight = float(value)
        
    @property
    def rimwidth(self) -> float:
        """
        The width of the crater rim in meters.
        
        Returns
        -------
        float
        """
        return self._rimwidth
    
    @rimwidth.setter
    def rimwidth(self, value: FloatLike) -> None:
        if not isinstance(value, FloatLike):
            raise TypeError("rimwidth must be of type FloatLike") 
        self._rimwidth = float(value)
        
    @property
    def peakheight(self) -> float:
        """
        The height of the central peak in meters.
        
        Returns
        -------
        float
        """
        return self._peakheight
    
    @peakheight.setter
    def peakheight(self, value: FloatLike) -> None:
        if not isinstance(value, FloatLike):
            raise TypeError("peakheight must be of type FloatLike") 
        self._peakheight = float(value)

    @property
    def floor_diameter(self) -> float:
        """
        The diameter of the crater floor in meters.
        
        Returns
        -------
        float
        """
        return self._floor_diameter
    
    @floor_diameter.setter
    def floor_diameter(self, value: FloatLike) -> None:
        if not isinstance(value, FloatLike):
            raise TypeError("floor_diameter must be of type FloatLike")
        self._floor_diameter = float(value)
        
    @property
    def floordepth(self) -> float:
        """
        Return the depth of the crater floor in m
        
        Returns
        -------
        float
        """
        return self._floordepth
    
    @floordepth.setter
    def floordepth(self, value: FloatLike) -> None:
        if not isinstance(value, FloatLike):
            raise TypeError("floordepth must be of type FloatLike")
        self._floordepth = float(value)
        
    @property
    def ejrim(self) -> float:
        """
        The thickness of ejecta at the rim in m.
        
        Returns
        -------
        float
        """
        return self._ejrim
    
    @ejrim.setter
    def ejrim(self, value: FloatLike) -> None:
        if not isinstance(value, FloatLike):
            raise TypeError("ejrim must be of type FloatLike")
        self._ejrim = float(value)
        
    @property
    def crater(self):
        return super().crater

    @crater.setter
    def crater(self, value):
        Morphology.crater.fset(self, value)
        self._set_morphology_config()

    @property
    def node_index(self):
        """
        The index of the node closest to the crater location.
        
        Returns
        -------
        int
        """
        return self._node_index
    
    @node_index.setter
    def node_index(self, value: int) -> None:
        if not isinstance(value, int):
            raise TypeError("node_index must be of type int")
        self._node_index = value

    @property
    def face_index(self):
        """
        The index of the face closest to the crater location.
        
        Returns
        -------
        int
        """
        return self._face_index
    
    @face_index.setter
    def face_index(self, value: int) -> None:
        if not isinstance(value, int):
            raise TypeError("face_index must be of type int")
        self._face_index = value

    @parameter
    def ejecta_truncation(self) -> float:
        """
        The radius at which the crater is truncated relative to the crater radius.
        
        Returns
        -------
        float or None
        """
        return self._ejecta_truncation 
    
    @ejecta_truncation.setter
    def ejecta_truncation(self, value: FloatLike | None): 
        if value is not None:
            if not isinstance(value, FloatLike):
                raise TypeError("truction_radius must be of type FloatLike")
            self._ejecta_truncation = float(value)
        else:
            self._ejecta_truncation = None
            
    @parameter
    def dorays(self) -> bool:
        """
        A flag to determine if the ray pattern should be used instead of the homogeneous ejecta blanket.
        
        Returns
        -------
        bool
        """
        return self._dorays
    
    @dorays.setter
    def dorays(self, value: bool) -> None:
        if not isinstance(value, bool):
            raise TypeError("dorays must be of type bool")
        self._dorays = value



import subprocess
import re
import setuptools_scm

v = setuptools_scm.get_version()

# Get version string from setuptools_scm
raw_version = subprocess.check_output(["python", "-m", "setuptools_scm"]).decode().strip()

match = re.match(r'^(\d+\.\d+\.\d+)([^\d].*)$', raw_version)
if match:
    version = f"{match.group(1)}-{match.group(2)}"
else:
    version = raw_version

cargo_path = "Cargo.toml"

# Read Cargo.toml as text
with open(cargo_path, "r", encoding="utf-8") as f:
    cargo_contents = f.read()

# Replace the version line
new_cargo_contents = re.sub(
    r'version\s*=\s*".*?"',
    f'version = "{version}"',
    cargo_contents,
    count=1,
)

# Write back
with open(cargo_path, "w", encoding="utf-8") as f:
    f.write(new_cargo_contents)
with open('cratermaker/_version.py', 'w') as f:
    f.write(f'__version__ = version = \"{version}\"\n')

print(f"Updated Cargo.toml to version: {version}")


from pathlib import Path
from typing import Any
from numpy.typing import NDArray
from cratermaker.components.surface import Surface
from cratermaker.utils.general_utils import parameter, format_large_units
from cratermaker.components.target import Target

@Surface.register("icosphere")
class IcosphereSurface(Surface):    
    """
    Create a uniform grid configuration using an icosphere. This is the most accurate and efficient way to create a uniform grid, but is limited to a few resolutions.
    
    Parameters
    ----------
    gridlevel : float
        The subdivision level of the icosphere. The number of faces is 20 * 4**gridlevel. The default gridlevel is 8.
    target : Target, optional
        The target body or name of a known target body for the impact simulation. 
    reset : bool, optional
        Flag to indicate whether to reset the surface. Default is True.
    regrid : bool, optional
        Flag to indicate whether to regrid the surface. Default is False.
    simdir : str | Path
        The main project simulation directory. Defaults to the current working directory if None.

    Returns
    -------
    IcosphereSurface
        An instance of the IcosphereSurface class initialized with the given pixel size. 
    """    
    
    def __init__(self, 
                 gridlevel: int = 8, 
                 target: Target | str | None = None,
                 reset: bool = False,
                 regrid: bool = False, 
                 simdir: str | Path | None = None,
                 **kwargs: Any):
        super().__init__(target=target, simdir=simdir, **kwargs)
        self.gridlevel = gridlevel
        self.load_from_files(reset=reset, regrid=regrid, **kwargs)

    def __repr__(self) -> str:
        base = super().__repr__()
        pix_mean = format_large_units(self.pix_mean, quantity="length")
        pix_std = format_large_units(self.pix_std, quantity="length")
        return (
            f"{base}\n"
            f"Grid Level: {self.gridlevel}\n" 
            f"Effective pixel size: {pix_mean} +/- {pix_std}"
        )   

    @property
    def _hashvars(self):
        """
        The variables used to generate the hash.
        """
        return [self._component_name, self.radius, self.gridlevel] 

    def generate_face_distribution(self, **kwargs: Any) -> NDArray:
        """
        Creates the points that define the mesh centers.
           
        Returns
        -------
        (3,n) ndarray of np.float64
            Array of points on a unit sphere.
        """ 
        from trimesh.creation import icosphere
       
        print(f"Generating a mesh with icosphere level {self.gridlevel}.")  
        mesh = icosphere(self.gridlevel)
        points = mesh.vertices.T
        return points
   
    @parameter
    def gridlevel(self) -> int:
        return self._gridlevel

    @gridlevel.setter
    def gridlevel(self, value: int) -> None:
        if value < 0:
            raise ValueError("Grid level must be a non-negative integer.")
        self._gridlevel = int(value)


    



from __future__ import annotations
from abc import abstractmethod
from typing import Any
from cratermaker.core.crater import Crater
from cratermaker.components.surface import Surface
from cratermaker.utils.component_utils import ComponentBase, import_components

class Morphology(ComponentBase):
    def __init__(self, crater : Crater | None = None, **kwargs: Any) -> None:
        """
        Initialize the Morphology class.

        Parameters
        ----------
        crater : Crater, optional
            The crater currently attached to the morphology model.
        **kwargs : Any
            Additional keyword arguments.

        Raises
        -------

        TypeError
            If the crater is not an instance of Crater.

        """
        super().__init__(**kwargs)
        object.__setattr__(self, "_crater" , crater)

    def __repr__(self) -> str:
        base = super().__repr__()
        if self.crater is None:
            return base
        return (
            f"{base}\n"
            f"{self.crater}"
        )

    @classmethod
    def maker(cls, 
             morphology: str | type[Morphology] | Morphology| None = None, 
             crater : Crater | None = None,
             **kwargs: Any) -> Morphology:
        """
        Initialize a component model with the given name or instance.

        Parameters
        ----------
        morphology : str or Morphology or None
            The name of the morphology model to use, or an instance of Morphology. If None, the default "simplemoon" is used.
        crater : Crater, optional
            The crater currently attached to the morphology model.
        **kwargs : Any
            Additional keyword arguments that are required for the specific morphology model being created.

        Returns
        -------
        component
            An instance of the specified component model.

        Raises
        ------
        KeyError
            If the specified morphology model name is not found in the registry.
        TypeError
            If the specified morphology model is not a string or a subclass of Morphology.
        """

        # Call the base class version of make and pass the morphology argument as the component argument
        if morphology is None:
            morphology = "simplemoon"
        morphology = super().maker(component=morphology, crater=crater, **kwargs)
        return morphology

    @abstractmethod
    def form_crater(self, 
                    surface: Surface,
                    crater: Crater | None = None,
                    **kwargs) -> None: ...    
    
    @property
    def crater(self):
        """
        The crater to be created.
        
        Returns
        -------
        Crater
        """ 
        return self._crater
    
    @crater.setter
    def crater(self, value):
        if value is not None and not isinstance(value, Crater):
            raise TypeError("crater must be an instance of Crater")
        self._crater = value


import_components(__name__, __path__, ignore_private=True)




"""
Plot the Neukum projectile CSFD
========================================

.. rubric:: By Austin Blevins and David Minton

In this example, we will plot the cumulative size-frequency distribution of projectiles for a 1 Gy age, which is derived from Ivanov, Neukum, and Wagner (2001) [#]_. 

.. [#] Ivanov, B.A., Neukum, G., Wagner, R., 2001. Size-Frequency Distributions of Planetary Impact Craters and Asteroids. In *Collisional Processes in the Solar System*, Springer Netherlands, Dordrecht, pp. 1-34.  https://doi.org/10.1007/978-94-010-0712-2_1 
"""

from cratermaker import Production
import matplotlib.pyplot as plt
import numpy as np
from matplotlib import ticker 
production = Production.maker("neukum", version="Projectile")
Dvals = np.logspace(-2, 7, num=1000)
Nvals = production.function(diameter=Dvals,age=1000.0)

fig, ax = plt.subplots(figsize=(4, 7))
Dvals *= 1e-3 # Convert from m to km
Nvals *= 1e6 # Convert from m^2 to km^2
x_min = Dvals[0]
x_max = Dvals[-1]
ax.set_xscale('log')
ax.set_yscale('log')
ax.set_ylabel('$\\mathregular{N_{>D}}$')
ax.set_xlabel('Projectile Diameter (km)')
ax.set_xlim(x_min, x_max)
ax.yaxis.set_major_locator(ticker.LogLocator(base=10.0, numticks=20))
ax.yaxis.set_minor_locator(ticker.LogLocator(base=10.0, subs=np.arange(2,10), numticks=100))
ax.xaxis.set_major_locator(ticker.LogLocator(base=10.0, numticks=20))
ax.xaxis.set_minor_locator(ticker.LogLocator(base=10.0, subs=np.arange(2,10), numticks=100))
ax.grid(True,which="minor",ls="-",lw=0.5,zorder=5)
ax.grid(True,which="major",ls="-",lw=1,zorder=10)
Dlo = production.sfd_range[0]*1e-3
Dhi = production.sfd_range[1]*1e-3
lo = Dvals < Dlo
hi = Dvals > Dhi
inrange = (Dvals >= Dlo) & (Dvals <= Dhi)
ax.plot(Dvals[inrange], Nvals[inrange], '-', color='black', linewidth=1.0, zorder=50)
ax.plot(Dvals[lo], Nvals[lo], '-.', color='orange', linewidth=2.0, zorder=50)
ax.plot(Dvals[hi], Nvals[hi], '-.', color='orange', linewidth=2.0, zorder=50)

plt.show()



import unittest
import cratermaker 
from cratermaker import Target
import tempfile
from pathlib import Path
import numpy as np
import xarray as xr
from cratermaker.constants import _COMBINED_DATA_FILE_NAME, _EXPORT_DIR
# This will suppress the warning issued by xarray starting in version 2023.12.0 about the change in the API regarding .dims
# The API change does not affect the functionality of the code, so we can safely ignore the warning
import warnings
warnings.filterwarnings("ignore",category=FutureWarning,module="xarray")
warnings.filterwarnings("ignore",category=FutureWarning,module="uxarray")

def warning_with_breakpoint(message, category, filename, lineno, file=None, line=None):
    print(f"{category} Warning in line {lineno} of {filename} : {message}")
    return
warnings.simplefilter("always")  # Always trigger the warnings
warnings.showwarning = warning_with_breakpoint

class TestSimulation(unittest.TestCase):
    
    def setUp(self):
        # Initialize a target and surface for testing
        target = Target(name="Moon")
        self.pix = target.radius / 10.0
        self.gridlevel = 5
        
    def test_simulation_defaults(self):
        with tempfile.TemporaryDirectory(ignore_cleanup_errors=True) as simdir:
            
            sim = cratermaker.Simulation(simdir=simdir, gridlevel=self.gridlevel)
            self.assertEqual(sim.target.name, "Moon")
        
    def test_simulation_save(self):
        with tempfile.TemporaryDirectory(ignore_cleanup_errors=True) as simdir:
            
            # Test basic save operation
            sim = cratermaker.Simulation(simdir=simdir,gridlevel=self.gridlevel)
            sim.save()
        
            # Test that variables are saved correctly
            sim.surface.set_elevation(1.0)
            np.testing.assert_array_equal(sim.surface.uxds["node_elevation"].values, np.ones(sim.surface.uxds.uxgrid.n_node)) 
            np.testing.assert_array_equal(sim.surface.uxds["face_elevation"].values, np.ones(sim.surface.uxds.uxgrid.n_face)) 
            
            sim.save()
            
            filename = Path(sim.data_dir) / _COMBINED_DATA_FILE_NAME.replace(".nc", f"{sim.interval_number:06d}.nc")
            self.assertTrue(filename.exists())
            with xr.open_dataset(filename) as ds:
                ds = ds.isel(time=-1)
                np.testing.assert_array_equal(ds["node_elevation"].values, np.ones(sim.surface.uxds.uxgrid.n_node))
                np.testing.assert_array_equal(ds["face_elevation"].values, np.ones(sim.surface.uxds.uxgrid.n_face))
        
            # Test saving combined data
            sim.save(combine_data_files=True)
            filename = _COMBINED_DATA_FILE_NAME
            
            filename = Path(sim.data_dir) / _COMBINED_DATA_FILE_NAME
            self.assertTrue(filename.exists())
            with xr.open_dataset(filename) as ds:
                ds = ds.isel(time=-1)
                np.testing.assert_array_equal(ds["node_elevation"].values, np.ones(sim.surface.uxds.uxgrid.n_node))
                np.testing.assert_array_equal(ds["face_elevation"].values, np.ones(sim.surface.uxds.uxgrid.n_face))
    
        return 
        
    def test_simulation_export_vtk(self):
        with tempfile.TemporaryDirectory(ignore_cleanup_errors=True) as simdir:
             
            sim = cratermaker.Simulation(simdir=simdir,gridlevel=self.gridlevel)
            # Test with default parameters
            default_out_dir = Path(sim.simdir) / _EXPORT_DIR
            expected_files = ["surface000000.vtp"]
            sim.export("vtk")
            self.assertTrue(Path(default_out_dir).is_dir()) 
            for f in expected_files:
                self.assertTrue((Path(default_out_dir / f).exists()))
            
    def test_emplace_crater(self):
        with tempfile.TemporaryDirectory(ignore_cleanup_errors=True) as simdir:
            cdiam = 2*self.pix
            sim = cratermaker.Simulation(simdir=simdir,gridlevel=self.gridlevel)
            sim.emplace_crater(final_diameter=cdiam)
            pdiam = sim.crater.projectile_diameter
            
            sim.emplace_crater(final_diameter=cdiam)
            sim.emplace_crater(projectile_diameter=pdiam)
        return
    
    def test_populate(self):
        with tempfile.TemporaryDirectory(ignore_cleanup_errors=True) as simdir:
            
            sim = cratermaker.Simulation(simdir=simdir,gridlevel=self.gridlevel)
            # Test that populate will work even if no craters are returned
            sim.populate(age=1e-6)
            
        sim.populate(age=3.8e3)
        return
    
    def test_run(self):
        with tempfile.TemporaryDirectory(ignore_cleanup_errors=True) as simdir:
            sim = cratermaker.Simulation(simdir=simdir,gridlevel=self.gridlevel)
            sim.run(age=1000)
            
            sim = cratermaker.Simulation(simdir=simdir,gridlevel=self.gridlevel)
            sim.run(age=1000, age_interval=100)

            # Test that the simulation doesn't fail when the age doesn't divide evenly by the age_interval
            sim = cratermaker.Simulation(simdir=simdir,gridlevel=self.gridlevel)
            sim.run(age=1010, age_interval=100)

    
    def test_invalid_run_args(self):
        with tempfile.TemporaryDirectory(ignore_cleanup_errors=True) as simdir:
            
            sim = cratermaker.Simulation(simdir=simdir,gridlevel=self.gridlevel)

            # Test case: Neither the age nor the diameter_number argument is provided
            with self.assertRaises(ValueError):
                sim.run()

            # Test case: Both the age and diameter_number arguments are provided
            with self.assertRaises(ValueError):
                sim.run(age=3.8e3, diameter_number=(300e3, 80))

            # Test case: Both the age_end and diameter_number_end arguments are provided
            with self.assertRaises(ValueError):
                sim.run(age_end=3.0e3, diameter_number_end=(300e3, 80))

            # Test case: The age argument is provided but is not a scalar
            with self.assertRaises(ValueError):
                sim.run(age=[3.8e3])

            # Test case: The age_end argument is provided but is not a scalar
            with self.assertRaises(ValueError):
                sim.run(age_end=[3.0e3])

            # Test case: The age_interval is provided but is not a positive scalar
            with self.assertRaises(ValueError):
                sim.run(age=3.8e3, age_interval=-100.0)

            # Test case: The age_interval provided is negative, or is greater than age - age_end
            with self.assertRaises(ValueError):
                sim.run(age=3.8e3, age_end=3.0e3, age_interval=1000.0)

            # Test case: The diameter_number argument is not a pair of values, or any of them are less than 0
            with self.assertRaises(ValueError):
                sim.run(diameter_number=(300e3, -80))

            # Test case: The diameter_number_end argument is not a pair of values, or any of them are less than 0
            with self.assertRaises(ValueError):
                sim.run(diameter_number_end=(300e3, -80))

            # Test case: The diameter_number_interval argument is not a pair of values, or any of them are less than 0
            with self.assertRaises(ValueError):
                sim.run(diameter_number_interval=(300e3, -80))

            # Test case: The age_interval and diameter_number_interval arguments are both provided
            with self.assertRaises(ValueError):
                sim.run(age=3.8e3, age_interval=100.0, diameter_number_interval=(300e3, 80))

            # Test case: The diameter_number_interval provided is negative, or is greater than diameter_number - diameter_number_end
            with self.assertRaises(ValueError):
                sim.run(diameter_number=(300e3, 80), diameter_number_end=(300e3, 30), diameter_number_interval=(300e3, 100))

            # Test case: The ninterval is provided but is not an integer or is less than 1
            with self.assertRaises(ValueError):
                sim.run(age=3.8e3, ninterval=0)

            # Test case: The ninterval is provided and either age_interval or diameter_number_interval is also provided
            with self.assertRaises(ValueError):
                sim.run(age=3.8e3, ninterval=100, age_interval=100.0)

        return

    def test_simulation_to_config(self):
        with tempfile.TemporaryDirectory(ignore_cleanup_errors=True) as simdir:
            # First simulation: no target passed, should default to "Moon"
            sim = cratermaker.Simulation(simdir=simdir, gridlevel=self.gridlevel)
            self.assertIsInstance(sim.target, Target)
            self.assertEqual(sim.target.name, "Moon")
            del sim

            # Second simulation: override target with "Mars"
            sim = cratermaker.Simulation(simdir=simdir, target="Mars", resume_old=True)
            self.assertEqual(sim.target.name, "Mars")
            del sim

            # Third simulation: no target passed, should read "Mars" from config
            sim = cratermaker.Simulation(simdir=simdir, resume_old=True)
            self.assertEqual(sim.target.name, "Mars")
            del sim
        return

if __name__ == '__main__':
    unittest.main()



import unittest
import tempfile
from cratermaker.components.morphology import Morphology
from cratermaker.components.target import Target
from cratermaker.components.surface import Surface
from cratermaker import Crater, Simulation

morphology_models = Morphology.available()
class TestMorphology(unittest.TestCase):
    def setUp(self):
        # Initialize a target and surface for testing
        self.target = Target(name="Moon")
        self.pix = self.target.radius / 10.0
        self.gridlevel = 5

        self.dummy_crater = Crater.maker(
            location=(0.0, 0.0),
            final_diameter=100000.0,
        )

    def test_model_registration(self):
        models = Morphology.available()
        self.assertIn("simplemoon", models)

    def test_model_instantiation(self):
        for model_name in morphology_models:
            morphology = Morphology.maker(model_name)
            self.assertIsInstance(morphology, object)
            self.assertEqual(morphology.name, model_name)
            morphology.crater = self.dummy_crater
            self.assertIs(morphology.crater, self.dummy_crater)

    def test_invalid_crater_type_raises(self):
        with self.assertRaises(TypeError):
            for model_name in morphology_models:
                morphology = Morphology.maker(model_name)
                morphology.crater = "not_a_crater"

    def test_form_crater_executes(self):
        with tempfile.TemporaryDirectory(ignore_cleanup_errors=True) as simdir:
            surface = Surface.maker(simdir=simdir, target=self.target, reset=True, gridlevel=self.gridlevel)
            for model_name in morphology_models:
                morphology = Morphology.maker(model_name)
                morphology.form_crater(surface, crater=self.dummy_crater)

    def testmake_morphology(self):
        # Test the make_morphology function
        for model_name in morphology_models:
            morphology = Morphology.maker(moprhology=model_name)
            self.assertIsInstance(morphology, Morphology)

    def test_crater_depth(self):
        # Tests that the surface elevations are expected

        final_diameter_list = [100e3, 200e3, 500e3, 1000e3]
        delta_vals = [0.4, 0.3, 0.3, 0.2]
            
        gridargs = {
            "icosphere": {
                "gridlevel" : 6
                },
            "arbitrary_resolution": {
                "pix": 10e3,
                },
            "hireslocal": {
                "pix": 1.0e3,
                "local_location": (0, 0), 
                "local_radius": 100e3, 
                }
            }

        with tempfile.TemporaryDirectory(ignore_cleanup_errors=True) as simdir:
            for name, args in gridargs.items():
                sim = Simulation(simdir=simdir, surface=name, **args)
                for final_diameter, delta in zip(final_diameter_list, delta_vals):
                    sim.surface.reset()
                    # verify that the surface is flat
                    self.assertAlmostEqual(sim.surface.node_elevation.min(), 0.0, delta=1e0)
                    self.assertAlmostEqual(sim.surface.face_elevation.min(), 0.0, delta=1e0)
                    self.assertAlmostEqual(sim.surface.node_elevation.max(), 0.0, delta=1e0)
                    self.assertAlmostEqual(sim.surface.face_elevation.max(), 0.0, delta=1e0)
                    
                    sim.emplace_crater(final_diameter=final_diameter, location=(0, 0))

                    # Verify that the crater depth and rim heights are close to the expected values
                    self.assertAlmostEqual(-sim.surface.node_elevation.min() / sim.morphology.floordepth, 1.0, delta=delta)
                    self.assertAlmostEqual(-sim.surface.face_elevation.min() / sim.morphology.floordepth, 1.0, delta=delta)
                    self.assertAlmostEqual(sim.surface.node_elevation.max() / sim.morphology.rimheight, 1.0, delta=2*delta)
                    self.assertAlmostEqual(sim.surface.face_elevation.max() / sim.morphology.rimheight, 1.0, delta=2*delta)

if __name__ == '__main__':
    unittest.main()


from numpy.random import Generator
from typing import Any
from cratermaker.components.projectile import Projectile
from warnings import warn
@Projectile.register("comets")
class CometProjectiles(Projectile):
    def __init__(self, 
                 target_name : str | None = None,
                 density : float | None = None,
                 rng: Generator | None = None,
                 rng_seed: int | None = None,
                 rng_state: dict | None = None,
                 **kwargs: Any):
        """
        An operations class for computing the projectile properties of a comet source population.

        Parameters
        ----------
        target_name : str
            The name of the target body for the impact.
        density : float
            The density of the projectile in kg/m^3. Default is 500.0 kg/m^3.
        rng : Generator | None
            A random number generator for Monte Carlo simulations. If None, a default generator will be used.
        rng_seed : int | None
            The random rng_seed for the simulation if rng is not provided. If None, a random rng_seed is used.    
        rng_state : dict, optional
            The state of the random number generator. If None, a new state is created.
        Notes
        -----
        The mean impact velocities for outer solar system bodies come from Table 1 of Zahnle et al. [1]_. For inner solar system bodies, from Table 2 of Borin et al. [2]_. 

        References
        ----------
        .. [1] Zahnle, K., Schenk, P., Levison, H., Dones, L., 2003. Cratering rates in the outer Solar System. Icarus 163, 263-289. https://doi.org/10.1016/S0019-1035(03)00048-4
        .. [2] Borin, P., Cremonese, G., Marzari, F., Lucchetti, A., 2017. Asteroidal and cometary dust flux in the inner solar system. A&A 605, A94. https://doi.org/10.1051/0004-6361/201730617


        **kwargs : Any
            Additional keyword arguments to be passed to internal functions.
        """

        # This model always samples velocities, angles, and directions, so override any values that may have been passed.
        if density is None:
            density = 500.0
        super().__init__(target_name=target_name, density=density, rng=rng, rng_seed=rng_seed, rng_state=rng_state, **kwargs)
        self.mean_velocity = self._set_mean_velocity()


    def _set_mean_velocity(self):
        """
        Sets the mean velocity of the projectile in m/s based on the target body.
        
        Returns
        -------
        float 
        """

        catalogue = {
            "Mercury" : 18400.0,
            "Venus" : 26300.0,
            "Earth" : 17670.0,
            "Moon" : 17670.0,
            "Mars" : 14210.0,
            "Metis": 59000.0,
            "Amalthea": 50000.0,
            "Thebe": 45000.0,
            "Io": 32000.0,
            "Europa": 26000.0,
            "Ganymede": 20000.0,
            "Callisto": 15000.0,
            "Himalia": 6100.0,
            "Prometheus": 32000.0,
            "Pandora": 31000.0,
            "Epimetheus": 30000.0,
            "Janus": 30000.0,
            "Mimas": 27000.0,
            "Enceladus": 24000.0,
            "Tethys": 21000.0,
            "Telesto": 21000.0,
            "Calypso": 21000.0,
            "Dione": 19000.0,
            "Helene": 19000.0,
            "Rhea": 16000.0,
            "Titan": 10500.0,
            "Hyperion": 9400.0,
            "Iapetus": 6100.0,
            "Phoebe": 3200.0,
            "Cordelia": 20000.0,
            "Ophelia": 19000.0,
            "Bianca": 18000.0,
            "Cressida": 18000.0,
            "Desdemona": 18000.0,
            "Juliet": 18000.0,
            "Portia": 18000.0,
            "Rosalind": 17000.0,
            "Belinda": 16000.0,
            "Puck": 15000.0,
            "Miranda": 12500.0,
            "Ariel": 10300.0,
            "Umbriel": 8700.0,
            "Titania": 6800.0,
            "Oberon": 5900.0,
            "Naiad": 22000.0,
            "Thalassa": 22000.0,
            "Despina": 21000.0,
            "Galatea": 20000.0,
            "Larissa": 18000.0,
            "Proteus": 14000.0,
            "Triton": 8200.0,
            "Nereid": 2800.0,
            "Pluto": 1900.0,
            "Charon": 1800.0,
            "KBO" : 1800.0,
        }
        if self.target_name in catalogue:
            pmv = float(catalogue[self.target_name])
        else:
            warn(f"Target {self.target_name} not found in known targets. Known targets include {list(catalogue.keys())}. Defaulting to KBO.")
            pmv = float(catalogue["KBO"])
        return pmv



# Default filenames and paths 
_CONFIG_FILE_NAME = "cratermaker.yaml"
_DATA_DIR = "surface_data"
_EXPORT_DIR = "export"
_COMBINED_DATA_FILE_NAME = "surface.nc"
_GRID_FILE_NAME = "grid.nc"
_CIRCLE_FILE_NAME = "circles.vtp"
_VTK_FILE_EXTENSION = "vtp"
_COMPONENT_NAMES = ["target", "scaling", "production", "morphology", "projectile", "surface"]

_SMALLFAC = 1.0e-5 # This is a factor used to determine the smallest length scale in the grid

# Optional: controlled public API
__all__ = []


from __future__ import annotations
from pathlib import Path
from abc import abstractmethod
import os
import shutil
import tempfile
import numpy as np
import uxarray as uxr
from uxarray import INT_FILL_VALUE, UxDataArray, UxDataset
import xarray as xr
from scipy.optimize import curve_fit, OptimizeWarning
from numpy.typing import NDArray, ArrayLike
from typing import Any
import hashlib
from cratermaker.constants import _GRID_FILE_NAME, _DATA_DIR, _SMALLFAC, _COMBINED_DATA_FILE_NAME 
from cratermaker.components.target import Target
from cratermaker.utils.custom_types import FloatLike, PairOfFloats
from cratermaker.utils.montecarlo_utils import get_random_location_on_face
from cratermaker.utils.component_utils import ComponentBase, import_components
from cratermaker._cratermaker import surface_functions
import warnings

class Surface(ComponentBase):
    _registry: dict[str, type[Surface]] = {}

    """
    This class is used for handling surface-related data and operations in the cratermaker project. It provides methods for 
    setting elevation data, calculating distances and bearings, and other surface-related computations.
    
    The Surface class extends UxDataset for the cratermaker project.
    
    Parameters
    ----------
    target : Target, optional
        The target body or name of a known target body for the impact simulation. 
    reset : bool, optional
        Flag to indicate whether to reset the surface. Default is the value of `regrid`
    regrid : bool, optional
        Flag to indicate whether to regrid the surface. Default is False. 
    simdir : str | Path
        The main project simulation directory. Defaults to the current working directory if None.
    **kwargs : Any
        Additional keyword arguments.
    """ 

    def __init__(self, 
                 target: Target | str | None = None,
                 simdir: str | Path | None = None,
                 **kwargs):

        object.__setattr__(self, "_target", None)
        object.__setattr__(self, "_uxds", None)
        object.__setattr__(self, "_pix_mean", None)
        object.__setattr__(self, "_pix_std", None)
        object.__setattr__(self, "_pix_min", None)
        object.__setattr__(self, "_pix_max", None)
        object.__setattr__(self, "_area", None)
        object.__setattr__(self, "_node_tree", None)
        object.__setattr__(self, "_face_tree", None)
        object.__setattr__(self, "_face_areas", None)
        object.__setattr__(self, "_smallest_length", None)
        super().__init__(simdir=simdir, **kwargs)
                       
        self._data_variable_init = {
            "node_elevation": {
                "units" : "m", 
                "long_name": "elevation of nodes", 
                "initial_value" : 0.0,
                "isfacedata" : False 
            },
            "face_elevation": {
                "units" : "m",
                "long_name": "elevation of faces",
                "initial_value" : 0.0,
                "isfacedata" : True
            },
            "ejecta_thickness": {
                "units" : "m",
                "long_name": "ejecta thickness",
                "initial_value" : 0.0,
                "isfacedata" : True
            },
            "ray_intensity": {
                "units" : "",
                "long_name": "ray intensity value",
                "initial_value" : 0.0,
                "isfacedata" : True
            },
        }

        self.target = Target.maker(target, **kwargs)

        return

    def __repr__(self) -> str:
        base = super().__repr__()
        return (
            f"{base}\n"
            f"Target: {self.target.name}\n"
            f"Grid File: {self.grid_file}"
        )

    @classmethod
    def maker(cls: Surface, 
             surface: str | Surface | None = None,
             target: Target | str | None = None,
             reset: bool = False,
             regrid: bool = False, 
             simdir: str | Path | None = None,
             **kwargs) -> Surface:
        """
        Factory method to create a Surface instance from a grid file.

        Parameters
        ----------
        surface : str or Surface, optional
            The name of the type of grid used for the surface. Default is "icosphere".
        target : Target, optional
            The target body or name of a known target body for the impact simulation. 
        reset : bool, optional
            Flag to indicate whether to reset the surface. Default is True.
        regrid : bool, optional
            Flag to indicate whether to regrid the surface. Default is False.
        simdir : str | Path
            The main project simulation directory. Defaults to the current working directory if None.
        **kwargs : Any
            Additional keyword arguments.

        Returns
        -------
        Surface
            An initialized Surface object.
        """
        
        if surface is None:
            surface = "icosphere"

        surface = super().maker(component=surface, 
                             target=target, 
                             reset = reset,
                             regrid = regrid,
                             simdir=simdir, 
                             **kwargs)
        return surface

    def load_from_files(self, reset: bool = False, **kwargs: Any) -> None:
        """
        Load the grid and data files into the surface object.
        This function loads the grid file and data files from the specified directory. If the grid file does not exist, it will attempt to create a new grid.
        If the data files do not exist, it will create an empty dataset. If reset is True, it will delete all data files except the grid file.

        Parameters
        ----------
        reset : bool, optional
            Flag to indicate whether to reset the surface. Default is False.
        """

        # Get the names of all data files in the data directory that are not the grid file
        regrid = self.regrid_if_needed(**kwargs)
        reset = reset or regrid

        data_file_list = list(self.data_dir.glob("*.nc"))
        if self.grid_file in data_file_list:
            data_file_list.remove(self.grid_file)

        # if data_file_list is empty, set reset to True
        reset = reset or not data_file_list

        # If reset is True, delete all data files except the grid file 
        if reset: 
            for f in data_file_list:
                f.unlink()  
            data_file_list = []
        
        try:
            with xr.open_dataset(self.grid_file) as uxgrid:
                uxgrid.load()
                if reset: # Create an empty dataset 
                    self._uxds = uxr.UxDataset()
                else: # Read data from from existing datafiles
                    with uxr.open_mfdataset(uxgrid, data_file_list, use_dual=False) as ds:
                        self._uxds = ds.isel(time=-1).load()
                self._uxds.uxgrid = uxr.Grid.from_dataset(uxgrid)
                self._uxgrid = uxgrid
        except:
            raise ValueError("Error loading grid and data files")
        
        if reset:
            self.reset(**kwargs)
        else:
            self.load_from_data()

        return

    def save_to_files(self, 
                      combine_data_files: bool = False,
                      interval_number: int = 0,
                      time_variables: dict | None = None,
                      *args, **kwargs) -> None:
        """
        Save the surface data to the specified directory. Each data variable is saved to a separate NetCDF file. If 'time_variables' is specified, then a one or more variables will be added to the dataset along the time dimension. If 'interval_number' is included as a key in `time_variables`, then this will be appended to the data file name.

        Parameters
        ----------
        combine_data_files : bool, optional
            If True, combine all data variables into a single NetCDF file, otherwise each variable will be saved to its own NetCDF file. Default is False.
        interval_number : int, optional
            Interval number to append to the data file name. Default is 0.
        time_variables : dict, optional
            Dictionary containing one or more variable name and value pairs. These will be added to the dataset along the time dimension. Default is None.
        """
        do_not_save = ["face_areas"]

        self.data_dir.mkdir(parents=True, exist_ok=True) 
        
        if time_variables is None:
            time_variables = {"elapsed_time":float(interval_number)}  
        else:
            if not isinstance(time_variables, dict):
                raise TypeError("time_variables must be a dictionary")
            
        # Variables that we do not want to save as they are computed at runtime     
        
        self.save_to_data()
        self.uxds.close()
        
        ds = self.uxds.expand_dims(dim="time").assign_coords({"time":[interval_number]})
        for k, v in time_variables.items():
            ds[k] = xr.DataArray(data=[v], name=k, dims=["time"], coords={"time":[interval_number]})
                    
        drop_vars = [k for k in ds.data_vars if k in do_not_save]
        if len(drop_vars) > 0:
            ds = ds.drop_vars(drop_vars)
            
        self._save_data(ds, interval_number, combine_data_files)

        return

    def generate_grid(self, **kwargs: Any) -> None: 
        """
        Generate a tessellated mesh of a sphere of based on the particular Surface component that is being used.
        """       
        self.grid_file.unlink(missing_ok=True)

        points = self.generate_face_distribution(**kwargs) 
        uxgrid = uxr.Grid.from_points(points, method="spherical_voronoi")
        uxgrid.attrs["_id"] = self._id
        with tempfile.NamedTemporaryFile(delete=False) as temp_file:
            uxgrid.to_xarray().to_netcdf(temp_file.name)
            temp_file.flush()
            os.fsync(temp_file.fileno())
            
        # Replace the original file only if writing succeeded
        shutil.move(temp_file.name,self.grid_file)      
        self._uxgrid = uxgrid

        regrid = self.regrid_if_needed(**kwargs)
        assert(not regrid)

        self._pix_mean, self._pix_std, self._pix_min, self._pix_max = self._compute_pix_size(uxgrid)

        return

    def regrid_if_needed(self, force : bool = False, **kwargs: Any) -> bool:
        """
        Check if the existing grid matches the desired parameters determine if regridding is necessary.

        This function checks if a grid file exists and matches the specified parameters based on a unique hash generated from these 
        parameters. If the grid does not exist or does not match the parameters it generates a new grid and returns True. 

        Parameters
        ----------
        force : bool, optional
            Flag to force regridding even if the grid file exists. Default is False.

        Returns
        -------
        bool
            A boolean indicating whether the grid should be regenerated. 
        """
    
        # Find out if the file exists, if it does't we'll need to make a new grid
        self.data_dir.mkdir(parents=True, exist_ok=True)
        regrid = force or not Path(self.grid_file).exists()

        if not regrid:
            try: 
                with xr.open_dataset(self.grid_file) as ds:
                    ds.load()
                    uxgrid = uxr.Grid.from_dataset(ds)
                    old_id = uxgrid.attrs.get("_id")
                    regrid = old_id != self._id
            except:
                regrid = True

        if regrid:
            print("Creating a new grid")
            try:
                self.generate_grid(**kwargs)
            except:
                raise RuntimeError("Failed to create a new grid")

        return regrid

    def reset(self, **kwargs: Any) -> None:
        """
        Reset the surface to its initial state.
        """
        for name, entry in self._data_variable_init.items():
            self.generate_data(name=name,
                               data=entry["initial_value"],
                               long_name=entry["long_name"],
                               units=entry["units"],
                               isfacedata=entry["isfacedata"],
                               save_to_file=True)
                               

        self.load_from_data()

    def load_from_data(self):
        self.face_lat = self.uxgrid.face_lat.values
        self.face_lon = self.uxgrid.face_lon.values
        self.node_lat = self.uxgrid.node_lat.values
        self.node_lon = self.uxgrid.node_lon.values
        self.node_elevation = self.uxds["node_elevation"].values
        self.face_elevation = self.uxds["face_elevation"].values
        self.ejecta_thickness = self.uxds["ejecta_thickness"].values
        self.ray_intensity = self.uxds["ray_intensity"].values
        
    def save_to_data(self):
        self.uxds["node_elevation"].values = self.node_elevation
        self.uxds["face_elevation"].values = self.face_elevation
        self.uxds["ejecta_thickness"].values = self.ejecta_thickness
        self.uxds["ray_intensity"].values = self.ray_intensity
    
    def full_view(self):
        return SurfaceView(self, slice(None), slice(None))

    @property
    def uxds(self) -> UxDataset:
        """
        The data associated with the surface. This is an instance of UxDataset.
        """
        return self._uxds

    @property
    def data_dir(self):
        """
        Directory for data files.
        """
        return self.simdir / _DATA_DIR

    @property
    def grid_file(self):
        """
        Path to the grid file.
        """
        return self.simdir / _DATA_DIR / _GRID_FILE_NAME

    @property
    def area(self):
        """
        Total surface area of the target body.
        """
        if self._area is None:
            self._area = self.face_areas.sum()
        return self._area

    @property
    def target(self):
        """
        The target body for the impact simulation. Set during initialization.
        """
        return self._target

    @target.setter
    def target(self, value):
        self._target = Target.maker(value)
        return 

    @property
    def radius(self):
        """
        Radius of the target body.
        """
        return self.target.radius

    @property
    def node_tree(self):
        if self._node_tree is None:
            self._node_tree = self.uxgrid.get_ball_tree("nodes", distance_metric="haversine", coordinate_system="spherical", reconstruct=True)
        
        return self._node_tree

    @property
    def face_tree(self):
        if self._face_tree is None:
            self._face_tree = self.uxgrid.get_ball_tree("face centers",  distance_metric="haversine", coordinate_system="spherical", reconstruct=True)
        
        return self._face_tree
    
    def generate_data(self,
                      name: str,
                      long_name: str | None = None,
                      units: str | None = None,
                      data: FloatLike | NDArray | None = None,
                      isfacedata: bool = True,
                      save_to_file: bool = False,
                      interval_number: int = 0,
                      combine_data_files: bool = False,
                    ) -> None:
        """
        Generate either a node or face data variable and optionally save it to a file. If the data variable already exists, it will be overwritten.

        Parameters
        ----------
        name : str
            Name of the data variable. This will also be used as the data file name.
        long_name : str, optional
            Long name of the data variable that will be saved as an attribute.
        units : str, optional
            Units of the data variable that will be saved as an attribute.
        data : scalar or array-like
            Data file to be saved. If data is a scalar, then the data file will be filled with that value. If data is an array, then the data file will be filled with the array values. The data array must have the same size as the number of faces or nodes in the grid.
        isfacedata : bool, optional
            Flag to indicate whether the data is face data or node data. Default is True.
        save_to_file: bool, optional
            Specify whether the data should be saved to a file. Default is False.
        interval_number : int, optional, default 0
            The interval number to use when saving the data to the data file.
        combine_data_files : bool, optional
            If True, combine the current data with the existing data for previous intervals in the data file. Default is False.
            
        Returns
        -------
        None
        """    
        with xr.open_dataset(self.grid_file) as ds:
            ds.load()
            uxgrid = uxr.Grid.from_dataset(ds)
        if long_name is None and units is None:
            attrs = None
        else:
            attrs = {}
            if long_name:
                attrs["long_name"] = long_name
            if units:
                attrs["units"] = units
                
        if isfacedata: 
            dims = ["n_face"]
            size = uxgrid.n_face
        else:
            dims = ["n_node"]
            size = uxgrid.n_node
    
        if data is None:
            data = np.zeros(size,dtype=np.float64) 
        elif np.isscalar(data):
            data = np.full(size,data)
        else:
            if data.size != size:
                raise ValueError("data must have the same size as the number of faces or nodes in the grid") 
        uxda = UxDataArray(
            data=data,
            dims=dims,
            attrs=attrs,
            name=name,
            uxgrid=uxgrid,
        ) 
         
        self._uxds[name] = uxda
        
        if save_to_file:
            self._save_data(uxda, interval_number, combine_data_files)
        return 
        
    def set_elevation(self, 
                    new_elev: NDArray[np.float64] | list[FloatLike] | None = None,
                    save_to_file: bool = False, 
                    combine_data_files: bool = False,
                    interval_number: int = 0,
                    ) -> None:
        """
        Set elevation data for the target's surface mesh.

        Parameters
        ----------
        new_elev : array_like, optional
            New elevation data to be set. If None, the elevation is set to zero. 
        save_to_file : bool, default False
            If True, save the elevation data to the elevation file.
        combine_data_files : bool, default False
            If True, combine the current data with the existing data for previous intervals in the data file.
        interval_number : int, default 0
            The interval number to use when saving the elevation data to the elevation file.
        """
        if new_elev is None or np.isscalar(new_elev):
            gen_node = True
            gen_face = True 
        elif new_elev.size == self.uxds.uxgrid.n_node:
            gen_node = True
            gen_face = False
        elif new_elev.size == self.uxds.uxgrid.n_face:
            gen_node = False
            gen_face = True
        else:
            gen_node = False
            gen_face = False
            raise ValueError("new_elev must be None, a scalar, or an array with the same size as the number of nodes in the grid")
    
        try:
            self.save_to_data()
        except AttributeError:
            pass
            
        if gen_node:
            self.generate_data(data=new_elev, 
                               name="node_elevation",
                               long_name="elevation of nodes",
                               units= "m",
                               isfacedata=False,
                               save_to_file=save_to_file,
                               combine_data_files=combine_data_files,
                               interval_number=interval_number
                              )   
        if gen_face:
            self.generate_data(data=new_elev, 
                               name="face_elevation",
                               long_name="elevation of faces",
                               units="m",
                               isfacedata=True,
                               save_to_file=save_to_file,
                               combine_data_files=combine_data_files,
                               interval_number=interval_number
                              )
        self.load_from_data()
        return 

    @staticmethod
    def calculate_haversine_distance(lon1: FloatLike, 
                    lat1: FloatLike, 
                    lon2: FloatLike, 
                    lat2: FloatLike,
                    radius: FloatLike = 1.0) -> float:
        """
        Calculate the great circle distance between two points on a sphere.

        Parameters
        ----------
        lon1 : FloatLike
            Longitude of the first point in radians.
        lat1 : FloatLike
            Latitude of the first point in radians.
        lon2 : FloatLike
            Longitude of the second point in radians.
        lat2 : FloatLike
            Latitude of the second point in radians.
        radius : FloatLike
            Radius of the sphere in meters.

        Returns
        -------
        float
            Great circle distance between the two points in meters.
        """
        # Calculate differences in coordinates
        dlon = lon2 - lon1
        dlat = lat2 - lat1

        # Haversine formula
        a = np.sin(dlat/2.0)**2 + np.cos(lat1) * np.cos(lat2) * np.sin(dlon/2.0)**2
        c = 2 * np.arcsin(np.sqrt(a))
        return radius * c
    
    def get_distance(self, 
                     view: 'SurfaceView',
                     location: tuple[float, float]) -> tuple[float, float]:
        """
        Computes the distances between nodes and faces and a given location.

        Parameters
        ----------
        location : tuple[float, float]
            tuple containing the longitude and latitude of the location in degrees.

        Returns
        -------
        UxDataArray
            DataArray of distances for each node in meters.
        UxDataArray
            DataArray of distances for each face in meters.
        """
        if len(location) == 1:
            location = location.item()
        lon1 = np.deg2rad(location[0])
        lat1 = np.deg2rad(location[1])
        node_lon2 = np.deg2rad(self.node_lon[view.node_indices])
        node_lat2 = np.deg2rad(self.node_lat[view.node_indices])
        face_lon2 = np.deg2rad(self.face_lon[view.face_indices])
        face_lat2 = np.deg2rad(self.face_lat[view.face_indices])
        return self.calculate_haversine_distance(lon1,lat1,node_lon2,node_lat2,self.radius), self.calculate_haversine_distance(lon1,lat1,face_lon2,face_lat2,self.radius) 

    @staticmethod
    def calculate_initial_bearing(lon1: FloatLike, 
                                lat1: FloatLike, 
                                lon2: FloatLike, 
                                lat2: FloatLike) -> float:
        """
        Calculate the initial bearing from one point to another on the surface of a sphere.

        Parameters
        ----------
        lon1 : FloatLike
            Longitude of the first point in radians.
        lat1 : FloatLike
            Latitude of the first point in radians.
        lon2 : FloatLike
            Longitude of the second point in radians.
        lat2 : FloatLike
            Latitude of the second point in radians.

        Returns
        -------
        float
            Initial bearing from the first point to the second point in radians.
        """
        # Calculate differences in coordinates
        dlon = np.mod(lon2 - lon1 + np.pi, 2 * np.pi) - np.pi

        # Haversine formula calculations
        x = np.sin(dlon) * np.cos(lat2)
        y = np.cos(lat1) * np.sin(lat2) - np.sin(lat1) * np.cos(lat2) * np.cos(dlon)
        initial_bearing = np.arctan2(x, y)

        # Normalize bearing to 0 to 2*pi
        initial_bearing = (initial_bearing + 2 * np.pi) % (2 * np.pi)

        return initial_bearing
    
    def get_initial_bearing(self, view: 'SurfaceView', location: tuple[float, float]) -> tuple[NDArray, NDArray]:
        """
        Computes the initial bearing between nodes and faces and a given location.

        Parameters
        ----------
        location : tuple[float, float]
            tuple containing the longitude and latitude of the location in degrees.

        Returns
        -------
        DataArray
            DataArray of initial bearings for each node in radians.
        DataArray
            DataArray of initial bearings for each face in radians.
        """
        if len(location) == 1:
            location = location.item()
        lon1 = np.deg2rad(location[0])
        lat1 = np.deg2rad(location[1])
        node_lon2 = np.deg2rad(self.node_lon[view.node_indices])
        node_lat2 = np.deg2rad(self.node_lat[view.node_indices])
        face_lon2 = np.deg2rad(self.face_lon[view.face_indices])
        face_lat2 = np.deg2rad(self.face_lat[view.face_indices])
        return (surface_functions.calculate_initial_bearing(lon1,lat1,node_lon2,node_lat2), 
                surface_functions.calculate_initial_bearing(lon1,lat1,face_lon2,face_lat2))
    
    def find_nearest_index(self, location):
        """
        Find the index of the nearest node and face to a given point.

        This method calculates the Haversine distance from the given point to each face in the grid,
        and returns the index of the face with the minimum distance.

        Parameters
        ----------
        location : tuple
            A tuple containing two elements: (longitude, latitude) in degrees.

        Returns
        -------
        int
            The index of the nearest node in the grid to the given point.
        int
            The index of the nearest face in the grid to the given point.

        Notes
        -----
        The method uses the ball tree query method that is included in the UxArray.Grid class.
        """          
        
        if len(location) == 1:
            location = location.item()
        coords = np.asarray(location)

        with warnings.catch_warnings():
            warnings.simplefilter("ignore", Warning)
            
            node_ind = self.node_tree.query(coords=coords, k=1, return_distance=False)
        
            face_ind = self.face_tree.query(coords=coords, k=1, return_distance=False)
        return node_ind.item(), face_ind.item()

    def get_reference_surface(self, 
                              view: 'SurfaceView',
                              face_crater_distance: NDArray, 
                              node_crater_distance: NDArray,
                              location: tuple[FloatLike, FloatLike], 
                              region_radius: float) -> NDArray[np.float64]:
        """
        Calculate the orientation of a hemispherical cap that represents the average surface within a given region.

        Parameters
        ----------
        location : tuple[float, float]
            tuple containing the longitude and latitude of the reference location in degrees.
        region_radius : float
            The radius of the region to compute the average over in meters.

        Returns
        -------
        center_vector : ndarray
            The vector pointing to the center of the cap from the sphere's center.
        """
        def sphere_function(coords, x_c, y_c, z_c, r):
            """
            Compute the sphere function.

            Parameters
            ----------
            coords : ndarray
                Array of x, y, and z coordinates.
            x_c, y_c, z_c : float
                Center coordinates of the sphere.
            r : float
                Radius of the sphere.

            Returns
            -------
            ndarray
                The values of the sphere function at the given coordinates.
            """
            x, y, z = coords.T
            return (x - x_c)**2 + (y - y_c)**2 + (z - z_c)**2 - r**2
   
        # Find cells within the crater radius
        faces_within_radius = face_crater_distance <= region_radius
        nodes_within_radius = node_crater_distance <= region_radius

        face_elevation = self.face_elevation[view.face_indices]
        node_elevation = self.node_elevation[view.node_indices]
        
        if np.sum(faces_within_radius) < 5 or not nodes_within_radius.any():
            return face_elevation, node_elevation
       
        face_grid = np.column_stack((self.uxds.uxgrid.face_x.values[view.face_indices], 
                               self.uxds.uxgrid.face_y.values[view.face_indices], 
                               self.uxds.uxgrid.face_z.values[view.face_indices]))
        region_faces = face_grid[faces_within_radius]
        region_elevation = face_elevation[faces_within_radius] / self.radius
        region_surf = self.elevation_to_cartesian(region_faces, region_elevation) 

        x, y, z = region_surf.T

        # Initial guess for the sphere center and radius
        guess_radius = 1.0 + region_elevation.mean()
        initial_guess = [0, 0, 0, guess_radius]  

        # Perform the curve fitting
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", OptimizeWarning)
            try:
                bounds =([-1.0, -1.0, -1.0, 0.5],[1.0, 1.0, 1.0, 2.0]) 
                popt, _  = curve_fit(sphere_function, region_surf[faces_within_radius], np.zeros_like(x[faces_within_radius]), p0=initial_guess, bounds=bounds)
            except:
                popt = initial_guess

        # Extract the fitted sphere center and radius
        reference_sphere_center = popt[:3]
        reference_sphere_radius = popt[3]
        
        def find_reference_elevations(coords):
            # Find the point along the original vector that intersects the sphere 
            f_vec = coords / self.radius      
            A = f_vec[:,0]**2 + f_vec[:,1]**2 + f_vec[:,2]**2
            B = -2 * (f_vec[:,0] * reference_sphere_center[0] + f_vec[:,1] * reference_sphere_center[1] + f_vec[:,2] * reference_sphere_center[2])
            C = np.dot(reference_sphere_center, reference_sphere_center) - reference_sphere_radius**2
            sqrt_term = B**2 - 4 * A * C
            valid = ~np.isnan(A) & (sqrt_term >= 0.0) 

            # Initialize t with default value
            t = np.full_like(A, 1.0)

            # Calculate square root only for valid terms
            sqrt_valid_term = np.sqrt(np.where(valid, sqrt_term, 0))

            # Apply the formula only where valid
            t = np.where(valid, (-B + sqrt_valid_term) / (2 * A), t)
            if np.any(t[valid] < 0):
                t = np.where(valid & (t < 0), (-B - sqrt_valid_term) / (2 * A), t)
                
            elevations = self.radius * (t * np.linalg.norm(f_vec, axis=1)  - 1)
            return elevations
        
        # Calculate the distances between the original face points and the intersection points
        reference_face_elevation = face_elevation
        reference_face_elevation[faces_within_radius] = find_reference_elevations(region_faces)
        
        # Now do the same thing to compute the nodal values 
        node_grid = np.column_stack((self.uxds.uxgrid.node_x.values[view.node_indices], 
                               self.uxds.uxgrid.node_y.values[view.node_indices], 
                               self.uxds.uxgrid.node_z.values[view.node_indices]))
        region_nodes = node_grid[nodes_within_radius]
        
        reference_node_elevation = node_elevation
        reference_node_elevation[nodes_within_radius] = find_reference_elevations(region_nodes)
        
        return reference_face_elevation, reference_node_elevation

    @staticmethod
    def elevation_to_cartesian(position: NDArray, 
                            elevation: NDArray
                            ) -> NDArray:
        
        
        runit = position / np.linalg.norm(position, axis=1, keepdims=True)
        
        return position + elevation[:, np.newaxis] * runit
   
    def extract_region(self,
                       location: tuple[FloatLike, FloatLike],
                       region_radius: FloatLike): 
        
        """
        Extract a regional grid based on a given location and radius.
        
        Parameters
        ----------
        location : tuple[float, float]
            tuple containing the longitude and latitude of the location in degrees.
        region_radius : float
            The radius of the region to extract in meters.
            
        Returns
        -------
        Surface
            A new Surface object containing the regional grid. 
            
        Notes
        -----
        Though not well documented, the mapping from the region grid back to the original grid can be done using the variables
        "subgrid_face_indices" and "subgrid_node_indices" in the region grid. For example, the following will extract a region
        surface, modify the face elevation, then insert the modified values back onto the original surface:
        
        .. code-block:: python    
              
            import xarray as xr
                
            region_surf = surface.extract_region(location, region_radius)
            region_surf['face_elevation'] = xr.full_like(region_surf['face_elevation'], 1.0)
            surface['face_elevation'].loc[{'n_face': region_surf.uxgrid._ds["subgrid_face_indices"]}] = region_surf['face_elevation']
            
        """ 
        
        region_angle = np.rad2deg(region_radius / self.radius)
        if len(location) == 1:
            location = location.item()
        coords = np.asarray(location)  

        ind = self.face_tree.query_radius(coords, region_angle)
        if len(ind) == 0:
            return None
        
        return SurfaceView(self, ind)
    
    def get_random_location_on_face(self, 
                                    face_index: int, 
                                    **kwargs
                                    ) -> float | tuple[float, float] | ArrayLike:
        """
        Generate a random coordinate within a given face of an ungridtyped mesh.

        Parameters
        ----------
        grid : uxarray.Grid
            The grid object containing the mesh information.
        face_index : int | NDArray[np.int64]
            The index or array of indices of the face within the grid to obtain the random sample.
        size : int or tuple of ints, optional
            The number of samples to generate. If size is None (the default), a single tuple is returned. If size is greater than 1, 
            then an array of tuples is returned. The default is 1.
                
        Returns
        -------
        (lon,lat) or ndarray[(lon,lat)] of given size
            A pair or array of pairs of longitude and latitude values in degrees.

        Notes
        -----
        This method is a wrapper for :func:`cratermaker.utils.montecarlo_utils.get_random_location_on_face`. 
        """
        
        return get_random_location_on_face(self.uxgrid, face_index, rng=self.rng, **kwargs)

    def _save_data(self,
                   ds: xr.Dataset | xr.DataArray,
                   interval_number: int = 0,
                   combine_data_files: bool = False) -> None:
        """
        Save the data to the specified directory. If `combine_data_files` is True, then all data variables are saved to a single NetCDF
        file. If False, then only the data variables for the current interval are saved to a NetCDF file with the interval number
        appended. 
        
        Parameters
        ----------
        ds : xr.Dataset or xr.DataArray
            The data to be saved.
        interval_number : int, Default is 0.
            Interval number to append to the data file name. Default is 0.
        combine_data_files : bool, Default is False.
            If True, combine all data variables into a single NetCDF file, otherwise each variable will be saved to its own NetCDF file. 
            
        Notes
        -----
        This function first saves to a temporary file and then moves that file to the final destination. This is done to avoid file 
        locking issues with NetCDF files.
        """
        if isinstance(ds, xr.DataArray):
            ds = ds.to_dataset()
            
        if "time" not in ds.dims:
            ds = ds.expand_dims(["time"])
        if "time" not in ds.coords:
            ds = ds.assign_coords({"time":[interval_number]})      
            
            
        with tempfile.TemporaryDirectory(ignore_cleanup_errors=True) as temp_dir:
            if combine_data_files:
                filename = _COMBINED_DATA_FILE_NAME
            else:
                filename = _COMBINED_DATA_FILE_NAME.replace(".nc", f"{interval_number:06d}.nc")
                
            data_file = self.data_dir / filename
            if data_file.exists():
                with xr.open_mfdataset(data_file) as ds_file:
                    ds_file = ds.merge(ds_file, compat="override")
                    ds_file.load()
            else:
                ds_file = ds    
                
            temp_file = Path(temp_dir) / filename
            
            comp = dict(zlib=True, complevel=9)
            encoding = {var: comp for var in ds_file.data_vars}
            ds_file.to_netcdf(temp_file, encoding=encoding)
            ds_file.close()     
            shutil.move(temp_file, data_file)

        return
    
    def __del__(self):
        try:
            if hasattr(self, "_uxds") and hasattr(self._uxds, "close"):
                if hasattr(self._uxds, "uxgrid") and hasattr(self._uxds.uxgrid, "_ds"):
                    self._uxds.uxgrid._ds.close()
                self._uxds.close()
        except Exception:
            pass
        try:
            if hasattr(self, "_grid") and hasattr(self._grid, "uxgrid") and hasattr(self._grid.uxgrid, "_ds"):
                self._grid.uxgrid._ds.close()
        except Exception:
            pass

    @abstractmethod
    def generate_face_distribution(self, **kwargs: Any) -> tuple[NDArray,NDArray,NDArray]: ...

    @property
    def _hashvars(self):
        """
        The variables used to generate the hash.
        """
        return [self._component_name, self._radius]
    
    @property
    def _id(self):
        """
        The hash id of the grid. This is used for determining if the grid needs to be regridded.
        """
        combined = ":".join(str(v) for v in self._hashvars)
        hash_object = hashlib.sha256(combined.encode())
        return hash_object.hexdigest()

    @staticmethod 
    def _distribute_points(distance: FloatLike,
                           radius: FloatLike=1.0, 
                           lon_range: PairOfFloats=(-180,180), 
                           lat_range: PairOfFloats=(-90,90)) -> NDArray:
        """
        Distributes points on a sphere using Deserno's algorithm (Deserno 2004).
            
        Parameters
        ----------
        distance : float
            Approximate distance between points, used to determine the number of points, where n = 1/distance**2 when distributed over the whole sphere.
        radius : float, optional
            Radius of the sphere. Default is 1.0
        lon_range : tuple, optional
            Range of longitudes in degrees. Default is (-180,180).
        lat_range : tuple, optional
            Range of latitudes in degrees. Default is (-90,90).
            
        Returns
        -------
        (3,n) ndarray of np.float64
            Array of cartesian points on the sphere.
            
        .. rubric:: References
        
        - Deserno, Markus., 2004. How to generate equidistributed points on the surface of a sphere. https://www.cmu.edu/biolphys/deserno/pdf/sphere_equi.pdf
        
        """
        def _sph2cart(theta, phi, r):
            """ 
            Converts spherical coordinates to Cartesian coordinates.
            
            Parameters
            ----------
            theta : float
                Inclination angle in radians.
            phi : float
                Azimuthal angle in radians.
            r : float
                Radius.
            """
            x = r * np.sin(theta) * np.cos(phi)
            y = r * np.sin(theta) * np.sin(phi)
            z = r * np.cos(theta)
            return x, y, z
        
        phi_range = np.deg2rad(lon_range) + np.pi 
        theta_range = np.deg2rad(lat_range) + np.pi/2 
        points = []
        
        n = int(1/distance**2)
        if n < 1: 
            return 
            
        a = 4 * np.pi / n
        d = np.sqrt(a)
        Mtheta = int(np.round(np.pi / d))
        dtheta = np.pi / Mtheta
        dphi = a / dtheta
        
        thetavals = np.pi * (np.arange(Mtheta) + 0.5) / Mtheta
        thetavals = thetavals[(thetavals >= theta_range[0]) & (thetavals < theta_range[1])]
        
        for theta in thetavals:
            Mphi = int(np.round(2 * np.pi * np.sin(theta) / dphi))
            phivals = 2 * np.pi * np.arange(Mphi) / Mphi
            phivals = phivals[(phivals >= phi_range[0]) & (phivals < phi_range[1])]
            for phi in phivals:
                points.append(_sph2cart(theta, phi, radius))
        if len(points) == 0:
            return
        
        points = np.array(points,dtype=np.float64)
        points = points.T

        return points    

    @property
    def uxgrid(self):
        """
        The grid object.
        """
        if self.uxds is not None:
            return self.uxds.uxgrid

    def _compute_pix_size(self, uxgrid: UxDataset | None = None) -> tuple[float, float]:
        """
        Compute the effective pixel size of the mesh based on the face areas.

        Parameters
        ----------
        uxgrid : UxDataset
            The grid object containing the mesh information. If not set, it will be retrieved self.uxds.uxgrid

        Returns
        -------
        tuple[float, float, float, float]
            The mean, standard deviation, minimum, and maximum of the pixel size in meters.
        """
        if uxgrid is None:
            if self.uxgrid is None:
                return None, None, None, None
            else:
                uxgrid = self.uxgrid
        face_areas = uxgrid.face_areas 
        face_sizes = np.sqrt(face_areas / (4 * np.pi))
        pix_mean = face_sizes.mean().item() * self.radius
        pix_std = face_sizes.std().item() * self.radius
        pix_min = face_sizes.min().item() * self.radius
        pix_max = face_sizes.max().item() * self.radius
        return float(pix_mean), float(pix_std), float(pix_min), float(pix_max)
    
    @property
    def pix_mean(self):
        """
        The mean pixel size of the mesh.
        """
        if self._pix_mean is None and self.uxgrid is not None:
            self._pix_mean, self._pix_std, self._pix_min, self._pix_max = self._compute_pix_size()
        return self._pix_mean
    
    @property
    def pix_std(self):
        """
        The standard deviation of the pixel size of the mesh.
        """
        if self._pix_std is None and self.uxgrid is not None:
            self._pix_mean, self._pix_std, self._pix_min = self._compute_pix_size()
        return self._pix_std
    
    @property
    def pix_min(self):
        """
        The minimum pixel size of the mesh.
        """
        if self._pix_min is None and self.uxgrid is not None:
            self._pix_mean, self._pix_std, self._pix_min = self._compute_pix_size()
        return self._pix_min
    
    @property
    def pix_max(self):
        """
        The maximum pixel size of the mesh.
        """
        if self._pix_max is None and self.uxgrid is not None:
            self._pix_mean, self._pix_std, self._pix_min, self._pix_max = self._compute_pix_size()
        return self._pix_max
    
    @property
    def gridtype(self):
        """
        The name of the grid type.
        """
        return self._component_name
    
    @property
    def face_areas(self):
        """
        The face areas of the mesh.
        """
        if self._face_areas is None:
            self._face_areas = self.uxgrid.face_areas.values * self.radius**2
        return self._face_areas
    
    @property
    def smallest_length(self):
        """
        The smallest length of the mesh.
        """
        if self._smallest_length is None:
            self._smallest_length = np.sqrt(np.min(self.face_areas)) * _SMALLFAC  
        return self._smallest_length


class SurfaceView:
    def __init__(self, 
                 surface: Surface, 
                 face_indices: NDArray | slice, 
                 node_indices: NDArray | slice | None = None):
        self.surface = surface
        self.face_indices = face_indices
        if node_indices is None:
            node_indices = np.unique(surface.uxds.uxgrid.face_node_connectivity.values[face_indices].ravel())
            node_indices = node_indices[node_indices != INT_FILL_VALUE]
        
        self.node_indices = node_indices






import_components(__name__, __path__, ignore_private=True)



"""
Sample a power law and lunar Neukum Production Function
========================================================

.. rubric:: By Austin Blevins and David Minton

In this example, we use the :func:`Production.sample` method to sample populations of craters from a production function. We will sample 1000 craters from each production function and repeat this 100 times. The population samples are then plotted against the expected value given by :func:`Production.function`
"""

from cratermaker import Production, Target
from cratermaker import Target
import matplotlib.pyplot as plt
import numpy as np
import matplotlib.ticker as ticker

fig, axs = plt.subplots(1, 2, figsize=(8, 7))
axes = dict(zip(['Power Law', 'NPF (Moon)'], axs))

production = {
    'Power Law': Production.maker("powerlaw"),
    'NPF (Moon)': Production.maker("neukum", version="Moon")
}

target = Target.maker("Moon")
area = 4 * np.pi * target.radius**2
age = 4100.0
x_min = 1e0
x_max = 1e5
y_min = 1e0
y_max = 1e4
diameter_range = (2e3, 10000e3) # Range of diameters to generate in m
nD = 1000
Dvals = np.logspace(np.log10(x_min), np.log10(x_max), num=nD)
Nevaluations = 100
for name,ax in axes.items():
    ax.title.set_text(name)
    ax.set_xscale('log')
    ax.set_yscale('log')
    ax.set_ylabel('$\\mathregular{N_{>D}}$')
    ax.set_xlabel('Diameter (km)')
    ax.set_xlim(x_min, x_max)
    ax.set_ylim(y_min, y_max)
    ax.yaxis.set_major_locator(ticker.LogLocator(base=10.0, numticks=20))
    ax.yaxis.set_minor_locator(ticker.LogLocator(base=10.0, subs=np.arange(2,10), numticks=100))
    ax.xaxis.set_major_locator(ticker.LogLocator(base=10.0, numticks=20))
    ax.xaxis.set_minor_locator(ticker.LogLocator(base=10.0, subs=np.arange(2,10), numticks=100))

    # Plot each sampled population Monte Carlo style
    for i in range(Nevaluations):
        Dsampled, _ = production[name].sample(age=age, diameter_range=diameter_range, area=area, return_age=False)
        Dsampled = np.sort(Dsampled)[::-1]
        Nsampled = range(1, len(Dsampled)+1)
        ax.plot(Dsampled*1e-3, Nsampled, '-', color='cornflowerblue', linewidth=2.0, zorder=50, alpha=0.2)

    # Plot the production function (the expected values)
    Nvals = production[name].function(diameter=Dvals*1e3, age=age)
    Nvals *= area # convert from per unit area to total number
    ax.plot(Dvals, Nvals, '-', color='black', linewidth=1.0, zorder=50)

plt.tick_params(axis='y', which='minor')
plt.tight_layout()
plt.show()



import numpy as np
from numpy.random import Generator
from cratermaker.components.production import Production
from cratermaker.utils.custom_types import FloatLike
from cratermaker.utils.general_utils import R_to_CSFD, parameter, format_large_units
from numpy.typing import ArrayLike
from collections.abc import Sequence
from typing import Any, Union

@Production.register("neukum")
class NeukumProduction(Production):
    """
    An operations class for computing the the Neukum production function for the Moon and Mars.

    Parameters
    ----------
    version : {"Moon", "Mars", "Projectile"}, optional
        The specific model to use for the production function. "Moon" and "Mars" are both crater production functions, and
        "Projectile" is a projectile function. Defaults to "Moon".
    rng : numpy.random.Generator | None
        A numpy random number generator. If None, a new generator is created using the rng_seed if it is provided.
    rng_seed : Any type allowed by the rng_seed argument of numpy.random.Generator, optional
        The rng_rng_seed for the RNG. If None, a new RNG is created.
    rng_state : dict, optional
        The state of the random number generator. If None, a new state is created.
    **kwargs : Any
        Additional keyword arguments. 
        
    Notes
    ----- 
    The CSFD is computed using the model of Ivanov, Neukum, and Hartmann (2001) for the Moon and Mars, with 
    minor changes. Notably, there is a typo in the chronology function (Eq. 5) of the original paper. The linear term in the paper
    is given as 8.38e-4. The value should be 10^(a0), and therefore the number given in the paper is based on the "Old"
    coefficients from Neukum (1983). The correct value is 10^(-3.0876) = 8.17e-4. We compute the value from the coefficients 
    in our implementation of the chronology function.       
    
    The Mars production function is based on Ivanov (2001). The projectile production function is from Ivanov, Neukum, and Wagner (2001).
   
    .. rubric:: References

    - Neukum, G., Ivanov, B.A., Hartmann, W.K., 2001. Cratering Records in the Inner Solar System in Relation to 
      the Lunar Reference System. *Space Science Reviews*, 96, 55-86. https://doi.org/10.1023/A:1011989004263
    - Ivanov, B.A., 2001. Mars/Moon Cratering Rate Ratio Estimates. *Space Science Reviews*, 96, 87-104. https://doi.org/10.1023/A:1011941121102
    - Ivanov, B.A., Neukum, G., Wagner, R., 2001. Size-Frequency Distributions of Planetary Impact Craters 
      and Asteroids. In *Collisional Processes in the Solar System*, Springer Netherlands, Dordrecht, pp. 1-34. 
      https://doi.org/10.1007/978-94-010-0712-2_1
    - Neukum, G., 1983. Meteorite bombardment and planetary surfaces dating. Habilitation Dissertation for Faculty Membership, University of Munich.

    """
    def __init__(self, 
                 version: str | None = None,
                 rng: Generator | None = None, 
                 rng_seed: int | None = None,
                 rng_state: dict | None = None,
                 **kwargs: Any):
        super().__init__(rng=rng, rng_seed=rng_seed, rng_state=rng_state, **kwargs) 

        self.version = version or "Moon"

    def __repr__(self) -> str:
        base = super().__repr__()
        timelo = format_large_units(self.valid_time[0], quantity="time")
        timehi = format_large_units(self.valid_time[1], quantity="time")
        dlo = format_large_units(self.sfd_range[0], quantity="length")
        dhi = format_large_units(self.sfd_range[1], quantity="length")
        return (
            f"{base}\n"
            f"Version: {self.version}\n"
            f"Valid Time Range: {timelo} - {timehi}\n"
            f"Valid Diameter Range: {dlo} - {dhi}"
        )

    def function(self,
             diameter: FloatLike | Sequence[FloatLike] | ArrayLike = 1.0,
             age: FloatLike | Sequence[FloatLike] | ArrayLike = 1.0,
             age_end: FloatLike | Sequence[FloatLike] | ArrayLike | None = None,
             check_valid_time: bool=True,
             **kwargs: Any,
             ) -> Union[FloatLike, ArrayLike]:
        """
        Return the cumulative size-frequency distribution of craters over a given age range and crater diameter.

        Parameters
        ----------
        diameter : FloatLike or numpy array
            Crater diameter(s) in units of meters to compute corresponding cumulative number density value.
        age : FloatLike or ArrayLike, default=1.0
            Age in the past in units of My relative to the present, which is used compute the cumulative SFD.
        age_end, FloatLike or ArrayLike, optional
            The ending age in units of My relative to the present, which is used to compute the cumulative SFD. The default is 0 (present day).
        check_valid_time : bool, optional (default=True)
            If True, return NaN for age values outside the valid age range

        Returns
        -------
        FloatLike or numpy array of FloatLike
            The cumulative number of craters per square meter greater than the input diameter that would be expected to form on a 
            surface over the given age range.
        """
        age, age_end = self._validate_age(age, age_end) 
        diameter, _ = self._validate_csfd(diameter=diameter)
        
        diameter_array = np.asarray(self.csfd(diameter))
        age_difference = np.asarray(self.chronology(age, check_valid_time)) - np.asarray(self.chronology(age_end, check_valid_time))

        if diameter_array.ndim > 0 and age_difference.ndim > 0:
            return diameter_array[:, None] * age_difference
        else:
            return diameter_array * age_difference
    
    def chronology(self,
             age: FloatLike | Sequence[FloatLike] | ArrayLike = 1000.0,
             check_valid_time: bool=True,
             **kwargs: Any,
             ) -> Union[FloatLike, ArrayLike]:
        """
        Returns the relative number of craters produced over a given age range. This implements the chronology function given in
        Eq. 5 of Ivanov, Neukum, and Hartmann (2001) SSR v. 96 pp. 55-86, but takes in the age argument in the Cratermaker unit 
        system of My instead of Gy. 

        Parameters
        ----------
        age : FloatLike or ArrayLike, default=1000.0
            Age in the past relative to the present day to compute cumulative SFD in units of My. 
        check_valid_time : bool, optional (default=True)
            If True, return NaN for age values outside the valid age range

        Returns
        -------
        FloatLike or numpy array of FloatLike
            The number of craters relative to the amount produced in the last 1 My.
            
        """     
        time_Gy = np.array(age) * 1e-3  # Convert age range from My to Gy ago for internal functions
        
        def _N1km(age: FloatLike | Sequence[FloatLike] | ArrayLike,
                check_valid_time:bool=True
                ) -> Union[FloatLike, ArrayLike]:
            """
            Return the cumulative number of 1 km craters as a function of age in Gy. This is a direct implementation of Eq. 5 in
            Ivanov, Neukum, and Hartmann (2001) SSR v. 96 pp. 55-86 (with corrected coefficient for the linear term).  

            Parameters
            ----------
            age : FloatLike or numpy array
                Time ago in units of Gy
            check_valid_time : bool, optional (default=True)
                If True, return NaN for age values outside the valid age range        

            Returns
            -------
            FloatLike or numpy array
                The number of craters per square kilometer greater than 1 km in diameter
            """
            N1 = self.Cexp * (np.exp(age/self.tau) - 1.0) + self.Clin * age
            if check_valid_time:
                if self.valid_time[0] is not None:
                    min_time = self.valid_time[0] * 1e-3
                    N1 = np.where(age >= min_time, N1, np.nan)
                if self.valid_time[1] is not None:
                    max_time = self.valid_time[1] * 1e-3
                    N1 = np.where(age <= max_time, N1, np.nan) 
            return N1.item() if np.isscalar(age) else N1
        
        N1_values = _N1km(age=time_Gy,check_valid_time=check_valid_time)*1e-6
        
        return  N1_values 

    @property
    def valid_versions(self) -> list[str]:
        """
        The valid versions of the production function. 
        
        Returns
        -------
        list
            The valid versions of the production function.
        """
        return ["Moon", "Mars", "Projectile"] 

    @property
    def generator_type(self) -> str:
        """
        Get the generator type of the production function.
        
        Returns
        -------
        str
            The generator type of the production function.
        """
        return self._generator_type
    
    @generator_type.setter
    def generator_type(self, value: str) -> None:
        """
        The generator type for this production function is determined by the version. 
        """
        raise NotImplementedError("The generator type cannot be changed for this production function.")

    @parameter
    def version(self) -> str:
        """
        Get the version of the production function.
        
        Returns
        -------
        str
            The version of the production function.
        """
        return self._version
    
    @version.setter
    def version(self, value: str) -> None:
        """
        Set the version of the production function.
        
        Parameters
        ----------
        value : str
            The version of the production function.
        """
        if value.title() not in self.valid_versions:
            raise ValueError(f"Invalid version '{value}'. Valid options are {self.valid_versions}")
        self._version = value.title()

        if self._version == "Projectile":
            self._generator_type = "projectile"
        else:
            self._generator_type = "crater"

    @property
    def sfd_tables(self) -> dict[str, list[float]]:
        """
        Get the table of coefficients for the production function.
        
        Returns
        -------
        dict
            The table of coefficients for the production function. 
        """
        return {"Moon" : [
                        -3.0876,
                        -3.557528,
                        +0.781027,
                        +1.021521,
                        -0.156012,
                        -0.444058,
                        +0.019977,
                        +0.086850,
                        -0.005874,
                        -0.006809,
                        +8.25e-4, 
                        +5.54e-5
                    ],
                "Mars" : [
                        -3.384, 
                        -3.197,
                        +1.257,
                        +0.7915,
                        -0.4861,
                        -0.3630,
                        +0.1016,
                        +6.756e-2,
                        -1.181e-2,
                        -4.753e-3,
                        +6.233e-4,
                        +5.805e-5
                    ],
                "Projectile" : [
                        0,
                        +1.375458,
                        +1.272521e-1,
                        -1.282166,
                        -3.074558e-1,
                        +4.149280e-1,
                        +1.910668e-1,
                        -4.260980e-2,
                        -3.976305e-2,
                        -3.180179e-3,
                        +2.799369e-3,
                        +6.892223e-4,
                        +2.614385e-6,
                        -1.416178e-5,
                        -1.191124e-6
                    ]
            }
        
    @property
    def sfd_coef(self) -> list[float]:
        """
        Get the table of coefficients for the production function.
        
        Returns
        -------
        list
            The table of SFD coefficients for the production function. 
        """
        return self.sfd_tables[self.version]
    
    @property
    def Clin(self) -> float:
        """
        Get the linear coefficient for the production function.
        
        Returns
        -------
        float
            The linear coefficient for the production function.
        """
        return 10**(self.sfd_coef[0])
    
    @property
    def Cexp(self) -> float:
        """
        Get the exponential coefficient for the production function.
        
        Returns
        -------
        float
            The exponential coefficient for the production function.
        """
        Cexp_moon = 5.44e-14
        if self.version == "Moon":
            return Cexp_moon
        else:
            Clin_moon = 10**(self.sfd_tables["Moon"][0])
            return Cexp_moon * self.Clin / Clin_moon 
    
    @property
    def tau(self) -> float:
        """
        Get the time constant for the production function.
        
        Returns
        -------
        float
            The time constant for the production function.
        """
        return 1.0 / 6.93

    @property
    def sfd_range(self) -> tuple[float, float]:
        """
        The range of diameters over which the SFD is valid. The range is given in m. 
        
        Returns
        -------
        tuple
            The lower and upper bounds of the SFD range in m.
        """
        sfd_range = {
                "Moon" : (10.0,1000.0e3),
                "Mars" : (15.0,362.0e3),
                "Projectile" : (0.1, 200.0e3) # Estimated based on Fig. 16 of Ivanov et al. (2001)
            }

        return sfd_range[self.version]
    
    @property
    def valid_time(self) -> tuple[float, float]:
        """
        The range of ages over which the production function is valid. The range is given in My. 
        
        Returns
        -------
        tuple
            The lower and upper bounds of the valid time range in My.
        """
        return (0,4500)

    
    def csfd(self,diameter: FloatLike | ArrayLike,) -> FloatLike | ArrayLike:
        """
        Return the cumulative size frequency distribution of craters at a given age relative to age = 1 My ago per m^2.

        Parameters
        ----------
        diameter : FloatLike or ArrayLike
            units of meter 
            
        Returns
        -------
        FloatLike or numpy array
           Cumulative number density of per square meter greater than the input diameter.
        """      
        # Convert m to km for internal functions
        Dkm_lo = self.sfd_range[0] * 1e-3
        Dkm_hi = self.sfd_range[1] * 1e-3

        def _extrapolate_sfd(side: str = "lo") -> Union[FloatLike, ArrayLike]:
            """
            Return the exponent, p, and and proportionality constant, A, for  the extrapolated 
            CSFD in the form N(D) = A * D**-p. 

            Parameters
            ----------
            side : str
                The side of the range to extrapolate. Valid values are "lo" and "hi"

            Returns
            -------
            A, p
            """    
            if side == "lo":
                Dkm = Dkm_lo
            elif side == "hi":
                Dkm = Dkm_hi
            else:
                raise ValueError("side must be 'lo' or 'hi'")
            p = _dNdD(Dkm)
            A = _CSFD(Dkm)
            return A, p


        def _dNdD(Dkm: FloatLike | Sequence[FloatLike] | ArrayLike) -> Union[FloatLike, ArrayLike]:
            """
            Return the derivative of the cumulative size-frequency distribution as a function of diameter. For diameter values outside 
            the range of the NPF, the derivative is extrapolated using a power law.

            Parameters
            ----------
            Dkm : FloatLike or numpy array
                diameter in units of km

            Returns
            -------
            FloatLike or numpy array
                The differential number of craters (dN/dD) per square kilometer greater than Dkm in diameter at age = 1 Gy ago.
            """        
            def _dNdD_scalar(Dkm): 
                dcoef = self.sfd_coef[1:]
                if Dkm < Dkm_lo:
                    A, p = _extrapolate_sfd(side="lo")
                    return A * (p / Dkm) * (Dkm / Dkm_lo) ** p 
                elif Dkm > Dkm_hi:
                    A, p = _extrapolate_sfd(side="hi")
                    return A * (p / Dkm) * (Dkm / Dkm_hi) ** p 
                else:
                    return sum(co * np.log10(Dkm) ** i for i, co in enumerate(dcoef))
            
            return _dNdD_scalar(Dkm) if np.isscalar(Dkm) else np.vectorize(_dNdD_scalar)(Dkm)

    
        def _CSFD(Dkm: FloatLike | Sequence[FloatLike] | ArrayLike) -> Union[FloatLike, ArrayLike]:
            """
            Return the cumulative size-frequency distribution at the reference age of 1 Gy ago. For diameter values outside 
            the range of the NPF, the CSFD is extrapolated using a power law.

            Parameters
            ----------
            Dkm : FloatLike or numpy array
                diameter in units of km

            Returns
            -------
            FloatLike or numpy array
                The number of craters per square kilometer greater than Dkm in diameter at age=1 Gy ago.
            """
            def _CSFD_scalar(Dkm):
                if Dkm < Dkm_lo:
                    A, p = _extrapolate_sfd(side="lo")
                    return A * (Dkm / Dkm_lo) ** p
                elif Dkm > Dkm_hi:
                    A, p = _extrapolate_sfd(side="hi")
                    p -= 2.0 # Steepen the upper branch of the SFD to prevent anomolously large craters from forming
                    return A * (Dkm / Dkm_hi) ** p
                else:
                    logCSFD = sum(co * np.log10(Dkm) ** i for i, co in enumerate(self.sfd_coef))
                    return 10 ** logCSFD
        
            return _CSFD_scalar(Dkm) if np.isscalar(Dkm) else np.vectorize(_CSFD_scalar)(Dkm)


        if np.any(diameter < 0.0):
            raise ValueError("diameter must be greater than or equal to 0.0")
              
        Dkm = diameter * 1e-3 # Convert m to km for internal functions

        if self.version == "Projectile":
            Ncumulative = R_to_CSFD(R=_CSFD, D=Dkm) * 2.94e-5 # This is a multiplication factor that gets the projectile CSFD to approximately match the lunar crater CSFD 
        else:
            Ncumulative = _CSFD(Dkm) 
            
        return Ncumulative * 1e3 # convert from Gy^-1 km^-2 to My^-1 m^-2    


if __name__ == "__main__":
    from scipy.optimize import curve_fit
    import matplotlib.pyplot as plt
    import matplotlib.ticker as ticker
    from cratermaker.components.target import Target
    
    def plot_npf_csfd():
        fig = plt.figure(1, figsize=(8, 7))
        ax = {'Moon': fig.add_subplot(121),
            'Mars': fig.add_subplot(122)}

        tvals = [0.01,1.0,4.0]
        x_min = 1e-3
        x_max = 1e4
        y_min = 1e-12
        y_max = 1e6
        nD = 1000
        Dvals = np.logspace(np.log10(x_min), np.log10(x_max), num=nD)
        for key in ax:
            production = NeukumProduction(version=key)
            ax[key].title.set_text(key)
            ax[key].set_xscale('log')
            ax[key].set_yscale('log')
            ax[key].set_ylabel('$\\mathregular{N_{>D} (km^{-2})}$')
            ax[key].set_xlabel('Diameter (km)')
            ax[key].set_xlim(x_min, x_max)
            ax[key].set_ylim(y_min, y_max)
            ax[key].yaxis.set_major_locator(ticker.LogLocator(base=10.0, numticks=20))
            ax[key].yaxis.set_minor_locator(ticker.LogLocator(base=10.0, subs=np.arange(2,10), numticks=100))
            ax[key].xaxis.set_major_locator(ticker.LogLocator(base=10.0, numticks=20))
            ax[key].xaxis.set_minor_locator(ticker.LogLocator(base=10.0, subs=np.arange(2,10), numticks=100))
            ax[key].grid(True,which="minor",ls="-",lw=0.5,zorder=5)
            ax[key].grid(True,which="major",ls="-",lw=1,zorder=10)
            Dkm_lo = production.sfd_range[0] * 1e-3
            Dkm_hi = production.sfd_range[1] * 1e-3
            inrange = (Dvals >= Dkm_lo) & (Dvals <= Dkm_hi)
            lo = Dvals < Dkm_lo
            hi = Dvals > Dkm_hi
            for t in tvals:
                Nvals = production.function(diameter=Dvals*1e3,age=t*1e3)
                Nvals *= 1e6 # convert from m^-2 to km^-2
                ax[key].plot(Dvals[inrange], Nvals[inrange], '-', color='black', linewidth=1.0, zorder=50)
                ax[key].plot(Dvals[lo], Nvals[lo], '-.', color='orange', linewidth=2.0, zorder=50)
                ax[key].plot(Dvals[hi], Nvals[hi], '-.', color='orange', linewidth=2.0, zorder=50)
                labeli = int(0.25*nD)
                ax[key].text(Dvals[labeli],3*Nvals[labeli],f"{t:.2f} ", ha="left", va="top",rotation=-72)

        plt.tick_params(axis='y', which='minor')
        plt.tight_layout()
        plt.show()

        
    def plot_npf_N1_vs_T():
        fig = plt.figure(1, figsize=(8, 4))
        ax = fig.add_subplot(111)
        ax.set_yscale('log')
        ax.set_ylabel('$\\mathregular{N(1) (km^{-2})}$')
        ax.set_xlabel('Time (Gy ago)')
        ax.set_xlim(4.5, 0)
        moon = NeukumProduction(version="Moon")
        mars = NeukumProduction(version="Mars")
        tvals = np.linspace(4.5, 0.0, num=1000)
        N1_moon = moon.function(diameter=1000.0, age=tvals*1e3)*1e6
        N1_mars = mars.function(diameter=1000.0, age=tvals*1e3)*1e6
        ax.plot(tvals, N1_moon, '-', color='dimgrey', linewidth=2.0, zorder=50, label="Moon")
        ax.plot(tvals, N1_mars, '-', color='orange', linewidth=2.0, zorder=50, label="Mars")
        ax.legend()
        plt.tight_layout()
        plt.show() 


    def plot_npf_fit():
        # Define the power-law function
        def power_law(D, N1_coef, slope):
            return N1_coef * D ** slope
       
        # Create the lunar crater and projectile production functions 
        crater_production = NeukumProduction(version="Moon")
        projectile_production = NeukumProduction(version="Projectile",impact_velocity_model = "Moon_MBA")

        # Make a population of sizes that spans the size of interest
        Dc = np.logspace(-1,2)
        Di = np.logspace(-2,1)

        # Compute the reference N values (N>D at 1 My)
        Nc = crater_production.function(diameter=Dc,age=1.0)
        Ni = projectile_production.function(diameter=Di,age=1.0)
       
        D1proj = 22.3 # Approximate diameter of crater made by a 1 m projectile using the default Cratermaker scaling relationships
         
        N1_c = crater_production.function(diameter = D1proj, age = 1.0)
        N1_p = projectile_production.function(diameter = 1.0, age = 1.0)        

        # Fit the power-law function to the crater data
        params_crater, _ = curve_fit(power_law, Dc, Nc)

        # Fit the power-law function to the projectile data
        params_projectile, _ = curve_fit(power_law, Di, Ni * N1_c / N1_p)

        # Extracting the parameters
        N1_coef_crater, slope_crater = params_crater
        N1_coef_projectile, slope_projectile = params_projectile

        # Output the results
        print("Crater Production Fit Parameters N1_coef =", N1_coef_crater, ", slope =", slope_crater)
        print("Projectile Production Fit Parameters N1_coef =", N1_coef_projectile, ", slope =", slope_projectile)
        
        # Create the fitted curve data
        fitted_curve_crater = power_law(Dc, *params_crater)
        fitted_curve_projectile = power_law(Di, *params_projectile)

        # Plotting the crater data
        plt.figure(figsize=(10, 5))
        plt.subplot(1, 2, 1)
        plt.loglog(Dc, Nc, 'o', label='Original Data (Crater)')
        plt.loglog(Dc, fitted_curve_crater, '-', label='Fitted Curve (Crater)')
        plt.xlabel('Crater Diameter (m)')
        plt.ylabel('N>D')
        plt.title('Crater Production')
        plt.legend()

        # Plotting the projectile data
        plt.subplot(1, 2, 2)
        plt.loglog(Di, Ni * N1_c/N1_p, 'o', label='Original Data (Projectile)')
        plt.loglog(Di, fitted_curve_projectile, '-', label='Fitted Curve (Projectile)')
        plt.xlabel('Projectile Diameter (m)')
        plt.ylabel('N>D')
        plt.title('Projectile Production')
        plt.legend()

        # Show the plot
        plt.tight_layout()
        plt.show()        


    def plot_npf_proj_csfd():
        fig = plt.figure(1, figsize=(4, 7))
        ax = fig.add_subplot(111)

        x_min = 1e-5
        x_max = 1e4
        y_min = 1e-7
        y_max = 1e13
        nD = 1000
        Dvals = np.logspace(np.log10(x_min), np.log10(x_max), num=nD)
        production = NeukumProduction(version="Projectile",impact_velocity_model = "Moon_MBA")
        ax.set_xscale('log')
        ax.set_yscale('log')
        ax.set_ylabel('$\\mathregular{N_{>D}}$')
        ax.set_xlabel('Projectile Diameter (km)')
        ax.set_xlim(x_min, x_max)
        #ax.set_ylim(y_min, y_max)
        ax.yaxis.set_major_locator(ticker.LogLocator(base=10.0, numticks=20))
        ax.yaxis.set_minor_locator(ticker.LogLocator(base=10.0, subs=np.arange(2,10), numticks=100))
        ax.xaxis.set_major_locator(ticker.LogLocator(base=10.0, numticks=20))
        ax.xaxis.set_minor_locator(ticker.LogLocator(base=10.0, subs=np.arange(2,10), numticks=100))
        ax.grid(True,which="minor",ls="-",lw=0.5,zorder=5)
        ax.grid(True,which="major",ls="-",lw=1,zorder=10)
        Dkm_lo = production.sfd_range[0] * 1e-3
        Dkm_hi = production.sfd_range[1] * 1e-3
        inrange = (Dvals >= Dkm_lo) & (Dvals <= Dkm_hi)
        lo = Dvals < Dkm_lo
        hi = Dvals > Dkm_hi
        t = 1.0
        Nvals = production.function(diameter=Dvals*1e3,age=t*1e3)
        Nvals *= 1e6 # convert from m^-2 to km^-2
        ax.plot(Dvals[inrange], Nvals[inrange], '-', color='black', linewidth=1.0, zorder=50)
        ax.plot(Dvals[lo], Nvals[lo], '-.', color='orange', linewidth=2.0, zorder=50)
        ax.plot(Dvals[hi], Nvals[hi], '-.', color='orange', linewidth=2.0, zorder=50)

        plt.tick_params(axis='y', which='minor')
        plt.tight_layout()
        plt.show()    
   
   
    def plot_sampled_csfd():
        from cratermaker.components.production.powerlaw import PowerLawProduction
        fig = plt.figure(1, figsize=(8, 7))
        ax = {'Power Law': fig.add_subplot(121),
            'NPF (Moon)': fig.add_subplot(122)}

        production = {
                    'Power Law': PowerLawProduction(impact_velocity_model = "Moon_MBA"),
                    'NPF (Moon)': NeukumProduction(version="Moon")
                    }
        
        target = Target("Moon")
        area = 4*np.pi*target.radius**2 
        age = 4100.0
        x_min = 1e0
        x_max = 1e5
        y_min = 1e0
        y_max = 1e4
        diameter_range = (2e3,10000e3) # Range of diameters to generate in m
        nD = 1000
        Dvals = np.logspace(np.log10(x_min), np.log10(x_max), num=nD)
        Nevaluations = 100
        for key in ax:
            ax[key].title.set_text(key)
            ax[key].set_xscale('log')
            ax[key].set_yscale('log')
            ax[key].set_ylabel('$\\mathregular{N_{>D}}$')
            ax[key].set_xlabel('Diameter (km)')
            ax[key].set_xlim(x_min, x_max)
            ax[key].set_ylim(y_min, y_max)
            ax[key].yaxis.set_major_locator(ticker.LogLocator(base=10.0, numticks=20))
            ax[key].yaxis.set_minor_locator(ticker.LogLocator(base=10.0, subs=np.arange(2,10), numticks=100))
            ax[key].xaxis.set_major_locator(ticker.LogLocator(base=10.0, numticks=20))
            ax[key].xaxis.set_minor_locator(ticker.LogLocator(base=10.0, subs=np.arange(2,10), numticks=100))
            # Plot the sampled values
            for i in range(Nevaluations):
                Dsampled, _ = production[key].sample(age=age, diameter_range=diameter_range, area=area, return_age=False)
                Dsampled = np.sort(Dsampled)[::-1]
                Nsampled = range(1,len(Dsampled)+1) 
                ax[key].plot(Dsampled*1e-3, Nsampled, '-', color='cornflowerblue', linewidth=2.0, zorder=50, alpha=0.2)
               
            # Plot the production function 
            Nvals = production[key].function(diameter=Dvals*1e3,age=age)
            Nvals *= area # convert from per unit area to total number
            ax[key].plot(Dvals, Nvals, '-', color='black', linewidth=3.0, zorder=50)

        plt.tick_params(axis='y', which='minor')
        plt.tight_layout()
        plt.show()
    
            
    # plot_npf_csfd()
    # plot_npf_N1_vs_T()
    # plot_npf_fit()    
    # plot_npf_proj_csfd()
    plot_sampled_csfd()




import unittest
import tempfile
import numpy as np
from pathlib import Path
from cratermaker import Target, Surface
from cratermaker.utils.montecarlo_utils import get_random_location
from cratermaker.utils.general_utils import normalize_coords

surfacetypes = Surface.available()

class TestSurface(unittest.TestCase):
    def setUp(self):
        # Initialize a target and surface for testing
        self.target = Target(name="Moon")
        self.pix = self.target.radius / 10.0
        self.gridlevel = 4
        
        return
    
    def test_initialize_surface(self):
        with tempfile.TemporaryDirectory(ignore_cleanup_errors=True) as simdir:
            # Initializing it first should run the mesh generator
            surface = Surface.maker(simdir=simdir, gridlevel=self.gridlevel, target=self.target, reset=True)
            del surface

            # Try initializing the surface again with the same parameters. This should find the existing grid file and load it 
            surface = Surface.maker(simdir=simdir, gridlevel=self.gridlevel, target=self.target, reset=False)

            # Test regridding if the parameters change
            n_face_orig = surface.uxgrid.n_face
            del surface

            surface = Surface.maker(simdir=simdir, gridlevel=self.gridlevel-1, target=self.target, reset=False)
            self.assertGreater(n_face_orig, surface.uxds.uxgrid.n_face)
            del surface

            # Test different target values
            surface = Surface.maker(simdir=simdir, gridlevel=self.gridlevel, target=Target(name="Mars"), reset=False)
            del surface

            surface = Surface.maker(simdir=simdir, gridlevel=self.gridlevel, target="Mercury", reset=False)
            del surface
            
            # Test bad values
            with self.assertRaises(TypeError):
                surface = Surface.maker(simdir=simdir, gridlevel=self.gridlevel, target=1, reset=False)
            with self.assertRaises(ValueError):
                surface = Surface.maker(simdir=simdir, gridlevel=self.gridlevel, target="Arrakis", reset=False)
            with self.assertRaises(ValueError):
                surface = Surface.maker(simdir=simdir, gridlevel=self.gridlevel, target=Target(name="Salusa Secundus"), reset=False)
        return


    def test_set_elevation(self):
        with tempfile.TemporaryDirectory(ignore_cleanup_errors=True) as simdir:
            surface = Surface.maker(simdir=simdir, gridlevel=self.gridlevel, target=self.target, reset=True)
            # Test with valid elevation data
            new_elev = np.random.rand(surface.uxds.uxgrid.n_node)  # Generate random elevation data
            surface.set_elevation(new_elev)

            # Test with invalid elevation data (wrong size)
            new_elev = np.random.rand(surface.uxds.uxgrid.n_node + 1)  # Incorrect size

            # Expect ValueError for incorrect size
            with self.assertRaises(ValueError):
                surface.set_elevation(new_elev)

            # Test setting elevation to None (should set to zero)
            surface.set_elevation(None)

            # Check if the elevation data is set to zero
            np.testing.assert_array_equal(surface.uxds['node_elevation'].values, np.zeros(surface.uxds.uxgrid.n_node))
            np.testing.assert_array_equal(surface.uxds['face_elevation'].values, np.zeros(surface.uxds.uxgrid.n_face))
            
        return
    
    
    def test_calculate_haversine_distance(self):
        # Example coordinates (lat/lon in radians)
        lon1, lat1 = np.radians(0), np.radians(0)  # Equator, prime meridian
        lon2, lat2 = np.radians(90), np.radians(0)  # 90 degrees East, equator
        radius = 6371000  # Earth's radius in meters

        # Known distance should be 1/4 the circumference of the Earth
        expected_distance = np.pi * radius / 2
        calculated_distance = Surface.calculate_haversine_distance(lon1, lat1, lon2, lat2, radius)

        # Compare the expected and calculated distances
        self.assertAlmostEqual(calculated_distance, expected_distance, places=1)


    def test_get_face_distance(self):
        with tempfile.TemporaryDirectory(ignore_cleanup_errors=True) as simdir:
            
            surface = Surface.maker(simdir=simdir, gridlevel=self.gridlevel, target=self.target, reset=True)
            
            location = get_random_location()[0]
            lon = location[0]
            lat = location[1]

            # Convert important points to radians
            north = (lon, lat + 90)
            south = (lon, lat - 90)
            antipode = (lon-180, -lat)
            
            north = normalize_coords(north)
            south = normalize_coords(south)
            antipode = normalize_coords(antipode)
            
            _,north_idx = surface.find_nearest_index(north)
            _,south_idx = surface.find_nearest_index(south)
            _,antipode_idx = surface.find_nearest_index(antipode)
            
            north_distance = surface.target.radius * np.pi / 2 
            south_distance = surface.target.radius * np.pi / 2
            antipode_distance = surface.target.radius * np.pi
            delta = 2*self.pix

            # Test distances
            _,distances = surface.get_distance(surface.full_view(), location)
            self.assertAlmostEqual(distances[north_idx], north_distance, delta=delta, msg=f"North face distance ratio: {distances[north_idx].item()/north_distance}")
            self.assertAlmostEqual(distances[south_idx], south_distance, delta=delta, msg=f"South face distance ratio: {distances[south_idx].item()/south_distance}")
            self.assertAlmostEqual(distances[antipode_idx], antipode_distance, delta=delta, msg=f"Antipode face distance ratio: {distances[antipode_idx].item()/antipode_distance}")
        
        
    def test_get_node_distance(self):
        with tempfile.TemporaryDirectory(ignore_cleanup_errors=True) as simdir:
            
            surface = Surface.maker(simdir=simdir, gridlevel=self.gridlevel, target=self.target, reset=True)
            
            location = get_random_location()[0]
            lon = location[0]
            lat = location[1]

            # Convert important points to radians
            north = (lon, lat + 90)
            south = (lon, lat - 90)
            antipode = (lon+180, -lat)
            
            north = normalize_coords(north)
            south = normalize_coords(south)
            antipode = normalize_coords(antipode)
            
            north_idx,_ = surface.find_nearest_index(north)
            south_idx,_ = surface.find_nearest_index(south)
            antipode_idx,_ = surface.find_nearest_index(antipode)

            north_distance = surface.target.radius * np.pi / 2 
            south_distance = surface.target.radius * np.pi / 2
            antipode_distance = surface.target.radius * np.pi
            delta = 2*self.pix

            # Test distances
            node_distances,_ = surface.get_distance(surface.full_view(), location)
            self.assertAlmostEqual(node_distances[north_idx], north_distance, delta=delta, msg=f"North node distance ratio: {node_distances[north_idx].item()/north_distance}")
            self.assertAlmostEqual(node_distances[south_idx], south_distance, delta=delta, msg=f"South node distance ratio: {node_distances[south_idx].item()/south_distance}")
            self.assertAlmostEqual(node_distances[antipode_idx], antipode_distance, delta=delta, msg=f"Antipode node distance ratio: {node_distances[antipode_idx].item()/antipode_distance}")
            
        
    def test_calculate_initial_bearing(self):
        # Example coordinates (lat/lon in radians)
        lon1, lat1 = np.radians(0), np.radians(0)  # Equator, prime meridian
        lon2, lat2 = np.radians(90), np.radians(0)  # 90 degrees East, equator

        # The bearing from (0, 0) to (90, 0) should be 90 degrees in radians
        expected_bearing = np.radians(90)
        calculated_bearing = Surface.calculate_initial_bearing(lon1, lat1, lon2, lat2)

        # Compare the expected and calculated bearings
        self.assertAlmostEqual(calculated_bearing, expected_bearing, places=1)
        
    # def test_get_random_on_face(self):
    #     # Tests that the random location is within the face we expect
    #     surface = Surface.maker(gridlevel=self.gridlevel*2, target=self.target, reset=True)
    #     n_per_face = 10
    #     for i in surface.n_face:
    #         original_face_index = i.values.item()
    #         for _ in range(n_per_face):
    #             location = surface.get_random_location_on_face(original_face_index)
    #             _, new_face_index = surface.find_nearest_index(location)
    #             self.assertEqual(original_face_index, new_face_index) 
            
    def test_face_surface_values(self):
        with tempfile.TemporaryDirectory(ignore_cleanup_errors=True) as simdir:
            
            # Tests that the face_surface generates the correct values
            surface = Surface.maker(simdir=simdir, gridlevel=self.gridlevel, target=self.target, reset=True) 
            total_area_1 = surface.uxgrid.calculate_total_face_area()
            total_area_2 = surface.face_areas.sum().item()
            ratio = np.sqrt(total_area_2/total_area_1) / self.target.radius
            self.assertAlmostEqual(ratio, 1.0, places=2)
            self.assertAlmostEqual(total_area_2/ (4*np.pi*self.target.radius**2), 1.0, places=2)



    def test_generate_grid(self):
            
        gridargs = {
            "icosphere": {
                "level" :self.gridlevel, 
                },
            "arbitrary_resolution": {
                "pix": self.pix, 
                },
            "hireslocal": {
                "pix": self.pix, 
                "local_location": (0, 0), 
                "local_radius": 100e3, 
                "superdomain_scale_factor": 10
                }
            }

        with tempfile.TemporaryDirectory(ignore_cleanup_errors=True) as simdir:
            for name, args in gridargs.items():
                surface = Surface.maker(name,simdir=simdir,**args)
                self.assertTrue(Path(surface.grid_file).exists())
        
        return

if __name__ == '__main__':
    unittest.main()



from pathlib import Path
import numpy as np
from typing import Any
from numpy.typing import NDArray
from cratermaker.utils.custom_types import FloatLike
from cratermaker.components.surface import Surface
from cratermaker.utils.general_utils import parameter, format_large_units
from cratermaker.components.target import Target

@Surface.register("arbitrary_resolution")
class ArbitraryResolutionSurface(Surface):
    """
    Create a uniform grid configuration with an arbitrary user-defined pixel size. This will not be as nice as the regular IcosphereSurface, but can be any resolution desired.
    
    Parameters
    ----------
    pix : float
        The approximate face size for the mesh in meters.
    target : Target, optional
        The target body or name of a known target body for the impact simulation. 
    reset : bool, optional
        Flag to indicate whether to reset the surface. Default is True.
    regrid : bool, optional
        Flag to indicate whether to regrid the surface. Default is False.
    simdir : str | Path
        The main project simulation directory. Defaults to the current working directory if None.

    Returns
    -------
    ArbitraryResolutionSurface
        An instance of the ArbitraryResolutionSurface class initialized with the given pixel size. 
    """    
    
    def __init__(self, 
                 pix: FloatLike | None = None, 
                 target: Target | str | None = None,
                 reset: bool = False,
                 regrid: bool = False, 
                 simdir: str | Path | None = None,
                 **kwargs: Any):
        super().__init__(target=target, simdir=simdir, **kwargs)
        self.pix = pix
        self.load_from_files(reset=reset, regrid=regrid, **kwargs)

    def __repr__(self) -> str:
        base = super().__repr__()
        pix_mean = format_large_units(self.pix_mean, quantity="length")
        pix_std = format_large_units(self.pix_std, quantity="length")
        return (
            f"{base}\n"
            f"Effective pixel size: {pix_mean} +/- {pix_std}"
        )           

    @property
    def _hashvars(self):
        """
        The variables used to generate the hash.
        """
        return [self._component_name, self.radius, self.pix]

    @parameter
    def pix(self):
        """
        The approximate face size for a cell of the mesh.
        """
        return self._pix

    @pix.setter
    def pix(self, value: FloatLike):
        if value is None:
            value= np.sqrt(4 * np.pi * self.radius**2) * 1e-3  # Default mesh scale that is somewhat comparable to a 1000x1000 CTEM grid
        elif not isinstance(value, FloatLike) or np.isnan(value) or np.isinf(value) or value <= 0:
            raise TypeError("pix must be a positive float")
        self._pix = float(value)

    def generate_face_distribution(self, **kwargs: Any) -> NDArray:
        """
        Creates the points that define the mesh centers.
           
        Returns
        -------
        (3,n) ndarray of np.float64
            Array of points on a unit sphere.
        
        """                

        print(f"Generating a mesh with uniformly distributed faces of size ~{self.pix} m.")
        points = self._distribute_points(distance=self.pix/self.radius) 
        points[:,0] = np.array([0,0,1])
        points[:,-1] = np.array([0,0,-1])
        return points
    




from __future__ import annotations
import numpy as np
from astropy.constants import G
from typing import Any
from cratermaker.utils.general_utils import _set_properties, _create_catalogue, parameter, format_large_units
from cratermaker.utils.custom_types import FloatLike
from cratermaker.utils.component_utils import ComponentBase, import_components

class Target(ComponentBase):
    """
    Represents the target body in a crater simulation.

    This class encapsulates the properties of the target that is impacted, including
    its material composition, size, and other relevant physical characteristics.
    """
       
    def __init__(self, 
                 name: str, 
                 radius: FloatLike | None = None, 
                 diameter: FloatLike | None = None,
                 mass: FloatLike | None = None, 
                 transition_scale_type: str | None = None,
                 material: str | None = None,
                 density: FloatLike | None = None,
                 **kwargs: Any,
                 ):
        """
        Initialize the target object, setting properties from the provided arguments.

        Parameters
        ----------
        name : str or None
            Name of the target body.
        radius : FloatLike or None
            Radius of the target body in km.
        diameter : FloatLike or None
            Diameter of the target body in km.
        mass : FloatLike or None
            Mass of the target body in kg.
        transition_scale_type : str or None
            Simple-to-complex transition scaling to use for the surface (either "silicate" or "ice").
        material : str or None
            Name of the material composition of the target body.
        density : FloatLike or None
            Volumetric density of the surface of the target body in kg/m^3.

        **kwargs : Any
            Additional keyword argumments that could be set by the user.

        Notes
        -----
        - The `radius` and `diameter` parameters are mutually exclusive. Only one of them should be provided.
        - Parameters set explicitly using keyword arguments will override those drawn from the catalogue.
        """    
        object.__setattr__(self, "_updating", False)   # guard against recursive updates
        super().__init__(**kwargs)
        object.__setattr__(self, "_name", None)
        object.__setattr__(self, "_radius", None)
        object.__setattr__(self, "_mass", None)
        object.__setattr__(self, "_transition_scale_type", None)
        object.__setattr__(self, "_material", None)
        object.__setattr__(self, "_density", None)
        object.__setattr__(self, "_catalogue", None)


        # ensure that only either diamter of radius is passed
        size_values_set = sum(x is not None for x in [diameter, radius])
        if size_values_set > 1:
            raise ValueError("Only one of diameter or radius may be set")
        if diameter is not None:
            radius = diameter / 2.0

        catalogue = kwargs.pop("catalogue", self.catalogue)

        # Set properties for the Target object based on the arguments passed to the function
        _set_properties(self, 
                        name=name, 
                        radius=radius, 
                        mass=mass, 
                        material=material,
                        catalogue=catalogue,
                        density=density,
                        transition_scale_type=transition_scale_type, 
                        **kwargs
                    )
        arg_check = sum(x is None for x in [self.name, self.diameter, self.mass, self.transition_scale_type])
        if arg_check > 0:
            raise ValueError("Invalid Target")

        if self._density is None:
            if self.material in self.density_catalogue:
                self._density = self.density_catalogue[self.material]


    def __repr__(self) -> str:
        diameter = format_large_units(self.diameter, quantity="length")
        escape_velocity = format_large_units(self.escape_velocity, quantity="velocity")
        return (
            f"<Target: {self.name}>\n"
            f"Material: {self.material}\n"
            f"Diameter: {diameter}\n"
            f"Mass: {self.mass:.2e} kg\n"
            f"Surface density: {self.density:.1f} kg/m³\n"
            f"Transition Type: {self.transition_scale_type}\n"
            f"Escape Velocity: {escape_velocity}\n"
            f"Gravity: {self.gravity:.3f} m/s²"
        )

    @parameter
    def radius(self) -> float | None:
        return self._radius

    @radius.setter
    def radius(self, value: FloatLike):
        if value is not None and value <= 0:
            raise ValueError("Radius must be positive")
        setattr(self, "_radius", float(value))

    @property
    def diameter(self) -> float | None:
        if self._radius is not None:
            return 2 * self._radius

    @property
    def mass(self) -> float | None:
        return self._mass

    @mass.setter
    def mass(self, value: FloatLike):
        if value is not None and value <= 0:
            raise ValueError("Mass must be positive")
        setattr(self, "_mass", float(value))

    @property
    def name(self):
        """
        The name of the target body.
        
        Returns
        -------
        str 
        """
        return self._name

    @name.setter
    def name(self, value):
        
        if not isinstance(value, str) and value is not None:
            raise TypeError("name must be a string or None")
        self._name = value

    @property
    def density_catalogue(self):
        """
        The target catalogue used for the target body.
        
        Returns
        -------
        dict 
        """
        return {"Water" : 1000.0,
                "Sand" : 1750.0,
                "Dry Soil" : 1500.0,
                "Wet Soil" : 2000.0,
                "Soft Rock" :2250.0, 
                "Hard Rock" : 2500.0,
                "Ice" : 900.0}   

    @parameter
    def material(self):
        """
        The name of the material composition of the target body.
        
        Returns
        -------
        str 
        """
        return self._material

    @material.setter
    def material(self, value):
        if not isinstance(value, str) and value is not None:
            raise TypeError("material must be a string or None")
        self._material = value

    @parameter
    def density(self):
        """
        The volumetric density of the surface of the target body in kg/m^3.
        
        Returns
        -------
        float 
        """
        return self._density
    
    @density.setter
    def density(self, value):
        if value is not None:
            if not isinstance(value, FloatLike):
                raise TypeError("density must be a numeric value or None")
            if value <= 0:
                raise ValueError("density must be a positive number")
            self._density = float(value)
        else:
            self._density = None
        return

    @property
    def catalogue(self):
        """
        The target catalogue used for the target body.
        
        Returns
        -------
        dict 
        """
        def _make_target_catalogue():
            # Define some built-in catalogue values for known solar system targets of interest
            target_properties = [
                "name",    "radius",   "mass",      "material", "transition_scale_type"
            ]
            # The catalogue was created with Swiftest
            target_values = [
                ("Mercury", 2439.40e3, 3.301001e+23, "Soft Rock", "silicate"),
                ("Venus", 6051.84e3, 4.867306e+24, "Hard Rock", "silicate"),
                ("Earth", 6371.01e3, 5.972168e+24, "Wet Soil", "silicate"),
                ("Moon", 1737.53e3, 7.345789e+22, "Soft Rock", "silicate"),
                ("Mars", 3389.92e3, 6.416909e+23, "Soft Rock", "silicate"),
                ("Phobos", 11.17e3, 1.080000e+16, "Soft Rock", "silicate"),
                ("Deimos", 6.30e3, 1.800000e+15, "Soft Rock", "silicate"),
                ("Ceres", 469.70e3, 9.383516e+20, "Soft Rock", "ice"),
                ("Vesta", 262.70e3, 2.590270e+20, "Soft Rock", "silicate"),
                ("Io", 1821.49e3, 8.929649e+22, "Hard Rock", "silicate"),
                ("Europa", 1560.80e3, 4.798574e+22, "Ice", "ice"),
                ("Ganymede", 2631.20e3, 1.481479e+23, "Ice", "ice"),
                ("Callisto", 2410.30e3, 1.075661e+23, "Ice", "ice"),
                ("Titan", 2575.50e3, 1.345181e+23, "Ice", "ice"),
                ("Rhea", 764.50e3, 2.306459e+21, "Ice", "ice"),
                ("Dione", 562.50e3, 1.095486e+21, "Ice", "ice"),
                ("Tethys", 536.30e3, 6.174430e+20, "Ice", "ice"),
                ("Enceladus", 252.30e3, 1.080318e+20, "Ice", "ice"),
                ("Mimas", 198.80e3, 3.750939e+19, "Ice", "ice"),
                ("Ariel", 578.90e3, 1.250019e+21, "Ice", "ice"),
                ("Umbriel", 584.70e3, 1.279535e+21, "Ice", "ice"),
                ("Titania", 788.90e3, 3.338178e+21, "Ice", "ice"),
                ("Oberon", 761.40e3, 3.076577e+21, "Ice", "ice"),
                ("Miranda", 235.70e3, 6.442623e+19, "Ice", "ice"),
                ("Triton", 1352.60e3, 2.140292e+22, "Ice", "ice"),
                ("Charon", 606.00e3, 1.589680e+21, "Ice", "ice"),
                ("Pluto", 1188.30e3, 1.302498e+22, "Ice", "ice"),
                ("Arrokoth", 9.13e3, 7.485000e+14, "Ice", "ice"),
            ]
            
            return _create_catalogue(target_properties, target_values)

        if self._catalogue is None:
            self._catalogue = _make_target_catalogue()
        return self._catalogue 

    @property
    def catalogue_key(self):
        """
        The key used to identify the property used as the key in a catalogue.
        """
        return "name"

    @property
    def transition_scale_type(self):
        """
        The type of simple-to-complex transition scaling used for the surface, either 'silicate' or 'ice'.
        
        Returns
        -------
        str 
        """
        return self._transition_scale_type

    @transition_scale_type.setter
    def transition_scale_type(self, value):
        valid_types = ["silicate", "ice"]
        if value not in valid_types and value is not None:
            raise ValueError(f"Invalid transition_scale_type: {value}. Must be one of {valid_types} or None")
        self._transition_scale_type = value

    # The following are computed properties based on radius and mass
    @property
    def escape_velocity(self) -> float:
        """
        Calculate the escape velocity for the target body in SI units.

        Returns
        -------
        float
            Escape velocity in m/s.
        """        
        return np.sqrt(2 * self.radius * self.gravity)
    
    @property
    def gravity(self) -> float | None:
        """
        Calculate the gravitational acceleration at the surface of the target body in SI units.

        Returns
        -------
        float
            Gravitational acceleration in m/s^2. 
        """
        return G.value*self.mass / (self.radius**2)  

    @classmethod
    def maker(cls : type[Target], 
             target : Target | str = "Moon", 
             radius: FloatLike | None = None, 
             diameter: FloatLike | None = None,
             mass: FloatLike | None = None, 
             transition_scale_type: str | None = None,
             material: str | None = None,
             density: FloatLike | None = None,
             **kwargs: Any) -> Target:
        """
        Initialize the target object, setting properties from the provided arguments.

        Parameters
        ----------
        target : str, Target, or None
            Name of the target body or a Target object.
        radius : FloatLike or None
            Radius of the target body in km.
        diameter : FloatLike or None
            Diameter of the target body in km.
        mass : FloatLike or None
            Mass of the target body in kg.
        transition_scale_type : str or None
            Simple-to-complex transition scaling to use for the surface (either "silicate" or "ice").
        material : str or None
            Name of the material composition of the target body.
        density : FloatLike or None
            Volumetric density of the surface of the target body in kg/m^3.
        **kwargs : Any
            Additional keyword argumments that could be set by the user.

        Notes
        -----
        - The `radius` and `diameter` parameters are mutually exclusive. Only one of them should be provided.
        - Parameters set explicitly using keyword arguments will override those drawn from the catalogue.
        """            

        if target is None:
            try: 
                target = cls(name="Moon", radius=radius, diameter=diameter, mass=mass, transition_scale_type=transition_scale_type, material=material, density=density, **kwargs)
            except:
                raise ValueError("Error initializing target.")
        elif isinstance(target, str):
            try:
                target = cls(name=target, radius=radius, diameter=diameter, mass=mass, transition_scale_type=transition_scale_type, material=material, density=density, **kwargs)
            except KeyError:
                raise ValueError(f"Target '{target}' not found in the catalogue. Please provide a valid target name.")
        elif not isinstance(target, Target):
            raise TypeError("target must be a string or a Target object")
        target._component_name = target.name
        
        return target


import_components(__name__, __path__, ignore_private=True)

 
